/*
 *  Released under the GNU GPL.  See http://www.gnu.org/licenses/gpl.txt
 *
 *  This program is part of TinyPLC
 *
 *  TinyPLC is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundatation; version 2 of the License.
 *
 *  TinyPLC is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 */
/*
 *  My thanks to my employer, Athena Controls (www.athenacontrols.com)
 *  for allowing me to open source this project.
 */
/*
 *  TinyPLC kernel
 *  Processor:   Atmel MEGA1281 or Atmel MEGA2561
 *  Compiler:    Codevision AVR C compiler, 1.24.8e
 *
 *  TinyPLC editor/compiler
 *  Processor:   PC-clone, VESA video (mode 0x105, 1024x768x256)
 *  Compiler:    DeSmet C, version 2.51 (PCC version 1.2)
 *
 *  Revision history:
 *  Programmer        Date       Comments
 *  ----------------  ---------  ---------------------------------------------
 *  William Couture   3/19/2006  Original code
 *  William Couture   9/13/2006  Modify for GPL release
 */

#define BROKEN_ADD_OR_BRANCH
/* #define SIMPLE_ADDROW */
#define SHOW_NODES

#define _PLC_C_
#include "plc.h"

#include "charset.h"
#include "bigchar.h"

char buff_512[1024];  /* OK, so I lied... */
int vesalist[512];
WORD vesa_mode;
int have_vesamode;
int have_vgamode;
int vesalistsize;

#define FREQUENCY 115200L   /* frequency IN to UART */

#define COMM_RETRY_MAX 3

#define QUEUE_SIZE 4096   /* chars in IN/OUT buf, keep this a power of 2 */
BYTE in_queue[QUEUE_SIZE];
WORD in_queue_head;
WORD in_queue_tail;
BYTE out_queue[QUEUE_SIZE];
WORD out_queue_head;
WORD out_queue_tail;

/* 8250 registers */

#define REC           0  /* UART receive reg */

#define XMIT          0  /* UART transmit reg */

#define INT_EN        1  /* UART int. enable reg */

#define INT_ID        2  /* UART int. ident. reg */

#define LINE_CONTROL  3  /* UART line control reg */

#define MODEM_CONTROL 4  /* UART modem control reg */

#define LINE_STATUS   5  /* UART line status reg */

#define MODEM_STATUS  6  /* UART modem status reg */

#define BAUD_LSB      0  /* UART baud divisor reg */

#define BAUD_MSB      1  /* UART baud divisor reg */
#define FCR_REG       2  /* FCR reg for 16550+ */

BYTE old_uart_setup[8];


#define NONE          0  /* Handshake param none */

#define HDW           1  /* Handshake param hardware */

#define XON           2  /* Handshake param software */



/* Interrupt enable register  */

#define RX_INT    0x01  /* Receive interrupt mask */

#define TBE_INT   0x02  /* Transmit buffer empty mask */

#define ERR_INT   0x04  /* Error interrupt mask */

#define RS_INT    0x08  /* Line interrupt mask */



/* Interrupt id register */

#define OUT2      0x08  /* Out 2 line */

#define DTR       0x01  /* DTR high */

#define RTS       0x02  /* RTS high */

#define CTS       0x10

#define DSR       0x20

#define XMTRDY    0x20

#define TXR          0  /*  Transmit register (WRITE) */


WORD commbase = 0x03f8;  /* default to COMM 1 */
WORD irqnum = 4;         /* default to COMM 1 */
DWORD baudrate = 9600;

WORD comm_int_seg, comm_int_ofs;  /* save area */

BYTE tempstr[128];

WORD videopage;


#define MENUSPACING (TEXTHEIGHT + 8)

#define MENULINE(n) (144 + n * MENUSPACING)
#define ENDLINE(n) (144 + TEXTHEIGHT + n * MENUSPACING)
#define MENUCOL1 0
#define ENDCOL1  (5 * TEXTWIDTH)
#define MENUCOL2 (TEXTWIDTH * 8)
#define ENDCOL2 (TEXTWIDTH * 13)

int havemouse, mouser, mousec;  /* Need SIGNED! */
int mousecolor = LTBLUE;
BYTE mouse_cursor[9] = { MOUSE_CURSOR, 0 };
BYTE mouse_overlay[9];
#define MOUSE_SAVE_ROWS TEXTHEIGHT
#define MOUSE_SAVE_COLS 8 * TEXTWIDTH
BYTE mouse_cursor_save[MOUSE_SAVE_ROWS][MOUSE_SAVE_COLS];
WORD mouse_object = NULL_OBJ;

WORD curpos_r, curpos_c;

BYTE null_str[] = { 0 };
BYTE three_char_inp[] = { START_NO_IN, CHAR_BOX, CHAR_BOX, CHAR_BOX, END_NO_IN, 0 };
BYTE two_char_inp[] = { START_NO_IN, CHAR_BOX, CHAR_BOX, END_NO_IN, VERT_BAR, 0 };
BYTE three_char_out[] = { START_NO_OUT, CHAR_BOX, CHAR_BOX, CHAR_BOX, END_NO_OUT, 0 };
BYTE two_char_out[] = { START_NO_OUT, CHAR_BOX, CHAR_BOX, END_NO_OUT, VERT_BAR, 0 };
BYTE no_input[] = { VERT_BAR, VERT_BAR, START_NO_IN, END_NO_IN, VERT_BAR, 0 };
BYTE nc_input[] = { VERT_BAR, VERT_BAR, START_NC_IN, END_NC_IN, VERT_BAR, 0 };
BYTE no_output[] = { VERT_BAR, VERT_BAR, START_NO_OUT, END_NO_OUT, VERT_BAR, 0 };
BYTE nc_output[] = { VERT_BAR, VERT_BAR, START_NC_OUT, END_NC_OUT, VERT_BAR, 0 };
BYTE lat_output[] = { START_NO_OUT, CHAR_BOX, CHAR_BOX, CHAR_BOX, END_NO_OUT, 0 };
BYTE or_branch[] = { BRANCH_BAR, START_BRANCH, END_BRANCH, BRANCH_BAR, BRANCH_BAR, 0 };
BYTE expand_branch[] = { BRANCH_BAR, START_BRANCH, END_BRANCH, EXPAND_BRANCH, BRANCH_BAR, 0 };
BYTE ladder_left[] = { LEFT_EDGE, 0 };
BYTE ladder_right[] = { RIGHT_EDGE, 0 };
BYTE rung_element[] = { VERT_BAR, VERT_BAR, VERT_BAR, VERT_BAR, VERT_BAR, VERT_BAR, VERT_BAR, VERT_BAR, 0 };
BYTE blank_element[] = "     ";
                     /* 12345 */
BYTE or_null[] = { VERT_BAR, 0 };
BYTE or_link[] = { OR_VERT_T, 0 };
BYTE or_vertical[] = { OR_VERT_BAR, 0 };
BYTE or_start[] = { OR_LEFT, 0 };
BYTE or_start_t[] = { OR_LEFT_T, 0 };
BYTE or_end[] = { OR_RIGHT, 0 };
BYTE or_end_t[] = { OR_RIGHT_T, 0 };
BYTE scroll_uparrow[] = { SCROLL_UPARROW, 0 };
BYTE scroll_downarrow[] = { SCROLL_DOWNARROW, 0 };
BYTE scroll_bar[] = { SCROLL_BAR, 0 };
BYTE scroll_box[] = { SCROLL_BOX, 0 };


void filesave(), fileload(), newladder(), download();
void onestep(), linestep(), run(), stop();
#ifdef SIMPLE_ADDROW
   void addrow(), delrow();
#else
   void appendrow();
#endif
void pickup(), quit();
void scrollladder();

void inp_init(), ninp_init(), out_init(), nout_init(), lout_init(), uout_init();
void add_init(), sub_init(), mul_init(), div_init(), mod_init();
void and_init(), or_init(), xor_init(), not_init(), neg_init();
void equ_init(), neq_init(), grt_init(), geq_init(), les_init(), leq_init();
void mov_init(), jmp_init(), osr_init(), ctt_init(), ctf_init();
void tmt_init(), tmf_init(), rtmt_init(), rtmf_init(), reset_init(), lbl_init();
void inv_init();
void inp_del(), ninp_del(), out_del(), nout_del(), lout_del(), uout_del();
void add_del(), sub_del(), mul_del(), div_del(), mod_del();
void and_del(), or_del(), xor_del(), not_del(), neg_del();
void equ_del(), neq_del(), grt_del(), geq_del(), les_del(), leq_del();
void mov_del(), jmp_del(), osr_del(), ctt_del(), ctf_del();
void tmt_del(), tmf_del(), rtmt_del(), rtmf_del(), reset_del(), lbl_del();
void inv_del();
void inp_draw(), ninp_draw(), out_draw(), nout_draw(), lout_draw(), uout_draw();
void add_draw(), sub_draw(), mul_draw(), div_draw(), mod_draw();
void and_draw(), or_draw(), xor_draw(), not_draw(), neg_draw();
void equ_draw(), neq_draw(), grt_draw(), geq_draw(), les_draw(), leq_draw();
void mov_draw(), jmp_draw(), osr_draw(), ctt_draw(), ctf_draw();
void tmt_draw(), tmf_draw(), rtmt_draw(), rtmf_draw(), reset_draw(), lbl_draw();
void inv_draw();
WORD inp_comp(), ninp_comp(), out_comp(), nout_comp(), lout_comp(), uout_comp();
WORD add_comp(), sub_comp(), mul_comp(), div_comp(), mod_comp();
WORD and_comp(), or_comp(), xor_comp(), not_comp(), neg_comp();
WORD equ_comp(), neq_comp(), grt_comp(), geq_comp(), les_comp(), leq_comp();
WORD mov_comp(), jmp_comp(), osr_comp(), ctt_comp(), ctf_comp();
WORD tmt_comp(), tmf_comp(), rtmt_comp(), rtmf_comp(), reset_comp(), lbl_comp();
WORD inv_comp();
void nullfn();

struct obj objs[NUMBER_OF_OBJECTS] =  /* indexed by enum objtypes */
{
   { blank_element, no_input,       inp_init,  inp_del,  inp_draw,  inp_comp  },
   { blank_element, nc_input,       ninp_init, ninp_del, ninp_draw, ninp_comp },
   { blank_element, no_output,      out_init,  out_del,  out_draw,  out_comp  },
   { blank_element, nc_output,      nout_init, nout_del, nout_draw, nout_comp },
   { " LAT ",       lat_output,     lout_init, lout_del, lout_draw, lout_comp },
   { " UNL ",       lat_output,     uout_init, uout_del, uout_draw, uout_comp },
   { " ADD ",       three_char_out, add_init,  add_del,  add_draw,  add_comp  },
   { " SUB ",       three_char_out, sub_init,  sub_del,  sub_draw,  sub_comp  },
   { " MUL ",       three_char_out, mul_init,  mul_del,  mul_draw,  mul_comp  },
   { " DIV ",       three_char_out, div_init,  div_del,  div_draw,  div_comp  },
   { " MOD ",       three_char_out, mod_init,  mod_del,  mod_draw,  mod_comp  },
   { " AND ",       three_char_out, and_init,  and_del,  and_draw,  and_comp  },
   { " OR ",        two_char_out,   or_init,   or_del,   or_draw,   or_comp   },
   { " XOR ",       three_char_out, xor_init,  xor_del,  xor_draw,  xor_comp  },
   { " NOT ",       three_char_out, not_init,  not_del,  not_draw,  not_comp  },
   { " NEG ",       three_char_out, neg_init,  neg_del,  neg_draw,  neg_comp  },
   { " EQ ",        two_char_inp,   equ_init,  equ_del,  equ_draw,  equ_comp  },
   { " NE ",        two_char_inp,   neq_init,  neq_del,  neq_draw,  neq_comp  },
   { " GT ",        two_char_inp,   grt_init,  grt_del,  grt_draw,  grt_comp  },
   { " GE ",        two_char_inp,   geq_init,  geq_del,  geq_draw,  geq_comp  },
   { " LT ",        two_char_inp,   les_init,  les_del,  les_draw,  les_comp  },
   { " LE ",        two_char_inp,   leq_init,  leq_del,  leq_draw,  leq_comp  },
   { " MOV ",       three_char_out, mov_init,  mov_del,  mov_draw,  mov_comp  },
   { " JMP ",       three_char_out, jmp_init,  jmp_del,  jmp_draw,  jmp_comp  },
   { " OSR ",       three_char_inp, osr_init,  osr_del,  osr_draw,  osr_comp  },
   { " CTT ",       three_char_inp, ctt_init,  ctt_del,  ctt_draw,  ctt_comp  },
   { " CTF ",       three_char_inp, ctf_init,  ctf_del,  ctf_draw,  ctf_comp  },
   { " TMT ",       three_char_inp, tmt_init,  tmt_del,  tmt_draw,  tmt_comp  },
   { " TMF ",       three_char_inp, tmf_init,  tmf_del,  tmf_draw,  tmf_comp  },
   { " RTT ",       three_char_inp, rtmt_init, rtmt_del, rtmt_draw, rtmt_comp },
   { " RTF ",       three_char_inp, rtmf_init, rtmf_del, rtmf_draw, rtmf_comp },
   { " RST ",       three_char_inp, reset_init,reset_del,reset_draw,reset_comp},
   { " INV ",       three_char_inp, inv_init,  inv_del,  inv_draw,  inv_comp },
   { " LBL ",       three_char_inp, lbl_init,  lbl_del,  lbl_draw,  lbl_comp  },
   { "Save",        null_str,       nullfn,    nullfn,    nullfn,   nullfn    },
   { "Load",        null_str,       nullfn,    nullfn,    nullfn,   nullfn    },
   { "New",         null_str,       nullfn,    nullfn,    nullfn,   nullfn    },
   { "Download"     null_str,       nullfn,    nullfn,    nullfn,   nullfn    },
   { "AddRow",      null_str,       nullfn,    nullfn,    nullfn,   nullfn    },
   { "DelRow",      null_str,       nullfn,    nullfn,    nullfn,   nullfn    },
   { "OneStep"      null_str,       nullfn,    nullfn,    nullfn,   nullfn    },
   { "LineStep",    null_str,       nullfn,    nullfn,    nullfn,   nullfn    },
   { "Run",         null_str,       nullfn,    nullfn,    nullfn,   nullfn    },
   { "Stop",        null_str,       nullfn,    nullfn,    nullfn,   nullfn    },
   { blank_element, or_branch,      nullfn,    nullfn,    nullfn,   nullfn    },
   { blank_element, expand_branch,  nullfn,    nullfn,    nullfn,   nullfn    },
   { "(c) Bill",    null_str,       nullfn,    nullfn,    nullfn,   nullfn    },
   { " Couture",    null_str,       nullfn,    nullfn,    nullfn,   nullfn    },
   { "    2006",    null_str,       nullfn,    nullfn,    nullfn,   nullfn    },
   { "V0.01A",      null_str,       nullfn,    nullfn,    nullfn,   nullfn    },
   { "Exit",        null_str,       nullfn,    nullfn,    nullfn,   nullfn    },
   { null_str,      null_str,       nullfn,    nullfn,    nullfn,   nullfn    }, /* ladder object */
   { null_str,      null_str,       nullfn,    nullfn,    nullfn,   nullfn    }, /* default mouse_object */
   { null_str,      null_str,       nullfn,    nullfn,    nullfn,   nullfn    }  /* scroll object */
};

struct element_info menu[] =
{
   {   0,     TEXTHEIGHT, MENUCOL1, MENUCOL1 + 4 * TEXTWIDTH, SAVE_OBJ, filesave, NULL },
   {   0,     TEXTHEIGHT, MENUCOL2, MENUCOL2 + 4 * TEXTWIDTH, LOAD_OBJ, fileload, NULL },
   {  32,  32+TEXTHEIGHT, MENUCOL1, MENUCOL1 + 3 * TEXTWIDTH, NEW_OBJ,  newladder, NULL },
   {  32,  32+TEXTHEIGHT, MENUCOL2, MENUCOL2 + 7 * TEXTWIDTH, DOWNLOAD_OBJ, download, NULL },
#ifdef SIMPLE_ADDROW
   {  56,  56+TEXTHEIGHT, MENUCOL1, MENUCOL1 + 6 * TEXTWIDTH, ADDROW_OBJ, addrow, NULL },
   {  56,  56+TEXTHEIGHT, MENUCOL2, MENUCOL2 + 6 * TEXTWIDTH, DELROW_OBJ, delrow, NULL },
#else
   {  56,  56+TEXTHEIGHT, MENUCOL1, MENUCOL1 + 6 * TEXTWIDTH, ADDROW_OBJ, pickup, appendrow },
   {  56,  56+TEXTHEIGHT, MENUCOL2, MENUCOL2 + 6 * TEXTWIDTH, DELROW_OBJ, pickup, NULL },
#endif
   {  88,  88+TEXTHEIGHT, MENUCOL1, MENUCOL1 + 7 * TEXTWIDTH, ONESTEP_OBJ,onestep, NULL },
   {  88,  88+TEXTHEIGHT, MENUCOL2, MENUCOL2 + 7 * TEXTWIDTH, LINESTEP_OBJ,linestep, NULL },
   { 112, 112+TEXTHEIGHT, MENUCOL1, MENUCOL1 + 3 * TEXTWIDTH, RUN_OBJ,   run, NULL },
   { 112, 112+TEXTHEIGHT, MENUCOL2, MENUCOL2 + 4 * TEXTWIDTH, STOP_OBJ,  stop, NULL },
   { MENULINE(0),  ENDLINE(0),  MENUCOL1, ENDCOL1, INP_OBJ,   pickup, NULL },
   { MENULINE(0),  ENDLINE(0),  MENUCOL2, ENDCOL2, NINP_OBJ,  pickup, NULL },
   { MENULINE(1),  ENDLINE(1),  MENUCOL1, ENDCOL1, OUT_OBJ,   pickup, NULL },
   { MENULINE(1),  ENDLINE(1),  MENUCOL2, ENDCOL2, NOUT_OBJ,  pickup, NULL },
   { MENULINE(2),  ENDLINE(2),  MENUCOL1, ENDCOL1, LOUT_OBJ,  pickup, NULL },
   { MENULINE(2),  ENDLINE(2),  MENUCOL2, ENDCOL2, UOUT_OBJ,  pickup, NULL },
   { MENULINE(3),  ENDLINE(3),  MENUCOL1, ENDCOL1, ADD_OBJ,   pickup, NULL },
   { MENULINE(3),  ENDLINE(3),  MENUCOL2, ENDCOL2, SUB_OBJ,   pickup, NULL },
   { MENULINE(4),  ENDLINE(4),  MENUCOL1, ENDCOL1, MUL_OBJ,   pickup, NULL },
   { MENULINE(4),  ENDLINE(4),  MENUCOL2, ENDCOL2, DIV_OBJ,   pickup, NULL },
   { MENULINE(5),  ENDLINE(5),  MENUCOL1, ENDCOL1, MOD_OBJ,   pickup, NULL },
   { MENULINE(5),  ENDLINE(5),  MENUCOL2, ENDCOL2, NEG_OBJ,   pickup, NULL },
   { MENULINE(6),  ENDLINE(6),  MENUCOL1, ENDCOL1, AND_OBJ,   pickup, NULL },
   { MENULINE(6),  ENDLINE(6),  MENUCOL2, ENDCOL2, OR_OBJ     pickup, NULL },
   { MENULINE(7),  ENDLINE(7),  MENUCOL1, ENDCOL1, XOR_OBJ,   pickup, NULL },
   { MENULINE(7),  ENDLINE(7),  MENUCOL2, ENDCOL2, NOT_OBJ,   pickup, NULL },
   { MENULINE(8),  ENDLINE(8),  MENUCOL1, ENDCOL1, EQU_OBJ,   pickup, NULL },
   { MENULINE(8),  ENDLINE(8),  MENUCOL2, ENDCOL2, NEQ_OBJ,   pickup, NULL },
   { MENULINE(9),  ENDLINE(9),  MENUCOL1, ENDCOL1, GRT_OBJ,   pickup, NULL },
   { MENULINE(9),  ENDLINE(9),  MENUCOL2, ENDCOL2, GEQ_OBJ,   pickup, NULL },
   { MENULINE(10), ENDLINE(10), MENUCOL1, ENDCOL1, LES_OBJ,   pickup, NULL },
   { MENULINE(10), ENDLINE(10), MENUCOL2, ENDCOL2, LEQ_OBJ    pickup, NULL },
   { MENULINE(11), ENDLINE(11), MENUCOL1, ENDCOL1, MOV_OBJ,   pickup, NULL },
   { MENULINE(11), ENDLINE(11), MENUCOL2, ENDCOL2, OSR_OBJ,   pickup, NULL },
   { MENULINE(12), ENDLINE(12), MENUCOL1, ENDCOL1, JMP_OBJ,   pickup, NULL },
   { MENULINE(12), ENDLINE(12), MENUCOL2, ENDCOL2, LBL_OBJ,   pickup, NULL },
   { MENULINE(13), ENDLINE(13), MENUCOL1, ENDCOL1, CTT_OBJ,   pickup, NULL },
   { MENULINE(13), ENDLINE(13), MENUCOL2, ENDCOL2, CTF_OBJ,   pickup, NULL },
   { MENULINE(14), ENDLINE(14), MENUCOL1, ENDCOL1, TMT_OBJ,   pickup, NULL },
   { MENULINE(14), ENDLINE(14), MENUCOL2, ENDCOL2, TMF_OBJ,   pickup, NULL },
   { MENULINE(15), ENDLINE(15), MENUCOL1, ENDCOL1, RTMT_OBJ,  pickup, NULL },
   { MENULINE(15), ENDLINE(15), MENUCOL2, ENDCOL2, RTMF_OBJ,  pickup, NULL },
   { MENULINE(16), ENDLINE(16), MENUCOL1, ENDCOL1, RESET_OBJ, pickup, NULL },
   { MENULINE(16), ENDLINE(16), MENUCOL2, ENDCOL2, INVERT_OBJ,pickup, NULL },
   { MENULINE(17), ENDLINE(17), MENUCOL1, ENDCOL1, ORBRANCH_OBJ, pickup, NULL },
   { MENULINE(17), ENDLINE(17), MENUCOL2, ENDCOL2, EXPAND_OBJ,pickup, NULL },
   { MENULINE(20), ENDLINE(20), MENUCOL1, ENDCOL1, COPYR1_OBJ,NULL, NULL },
   { MENULINE(20), ENDLINE(20), MENUCOL2, ENDCOL2, COPYR2_OBJ,NULL, NULL },
   { MENULINE(21), ENDLINE(21), MENUCOL1, ENDCOL1, COPYR3_OBJ,NULL, NULL },
   { MENULINE(22), ENDLINE(22), MENUCOL1, ENDCOL1, VERSION_OBJ,NULL, NULL },
   { MENULINE(25), ENDLINE(25), MENUCOL1, ENDCOL1, EXIT_OBJ,  quit, NULL },
   { SCROLL_TOP,   SCROLL_BOTTOM,SCROLL_LEFTCOL, SCROLL_RIGHTCOL, SCROLL_OBJ, scrollladder, NULL }
};
#define NUM_MENU_ELEMENTS 52
WORD screen_index = NUM_MENU_ELEMENTS;
struct element_info *screen_list[SCREEN_MAX];

WORD dirty = FALSE;
WORD exitflag = FALSE;
WORD quickexit = FALSE;
/* WORD quickexit = TRUE; */  /* default for testing */

void ladder_node_add();
void ladder_node_del();
struct ladder_rung rungs[MAX_RUNGS];
WORD used_rungs[MAX_RUNGS];
WORD rung_list[MAX_RUNGS+1];
WORD ladder_size = 1;

#define FILEVERSION_LEN 16
/* BYTE fileversion[FILEVERSION_LEN] = "Athena 0.00A"; */
                                  /* 0123456789ab */
BYTE fileversion[FILEVERSION_LEN] = "TinyPLC 0.01A";
                                  /* 0123456789abc */

DWORD modbus_dataregs[PLC_MEMORY_LOCS];
BYTE program[MAX_PROGRAM];
WORD program_len;
WORD label_addresses[NUM_LABELS];

WORD scroll_baserow;

WORD is_littleendian = TRUE;

WORD peekw(), get_timer();

main(argc,argv)
   int argc;
   char *argv[];
   {
   WORD i, j;

   get_vesa_mode();
   havemouse = init_mouse();
   if (!vesa_mode)
      {
      printf("Error:  suitable VESA mode not found!\n");
      exit(1);
      }
   if (!havemouse)
      {
      printf("No mouse available!\n");
      exit(1);
      }

   for (i = 1; i < argc; i++)
      {
      if (argv[i][0] == '-')
         {
         switch(toupper(argv[i][1]))
            {
            case 'B':  /* Baudrate */
               if (!argv[i][2])
                  help();
               sscanf(&argv[i][2],"%ld",&baudrate);
               break;
            case 'Q':  /* quick exit */
               quickexit = TRUE;
               break;
            case 'E':  /* big-endian target */
               is_littleendian = FALSE;
               break;
            default:
               help();  /* this does not return */
               break;
            }
         for (j = i; j < argc-1; j++)
            argv[j] = argv[j+1];
         argc--;
         i--;  /* remove this cmdline parameter */
         }
      }
   if (argc > 1)
      {
      switch(argv[1][0])
         {
         case '1':
            commbase = 0x03f8;
            irqnum = 4;
            break;
         case '2':
            commbase = 0x02f8;
            irqnum = 3;
            break;
         case '3':
            commbase = 0x03e8;
            irqnum = 4;
            break;
         case '4':
            commbase = 0x02e8;
            irqnum = 3;
            break;
         default:
            help();  /* this does not return */
            break;
         }
      }
   take_comm_interrupt();
   init_comm_port();
   vesamode(vesa_mode);

   dirty = FALSE;
   newladder();
   init_mouse_cursor();

   while (!exitflag)
      {
      update_mouse();
      if (csts())
         if (ci() == ESC)
            {
            if (!quickexit)
               cleanup_dirty();
            exitflag = TRUE;
            }
      }

   videomode(MODE_STANDARD_TEXT);
   reset_comm_port();
   reset_comm_interrupt();
   exit(0);
   }

WORD beep()
   {
   putchar('\007');
   return(FALSE);
   }

WORD mouseover_menuindex()
   {
   WORD i, menuindex;

   menuindex = NULL_OBJECT;
   for (i = 0; i < screen_index; i++)
      {
      if (screen_list[i].objindex == LADDER_OBJ)
         {
         if ((mouser+scroll_baserow >= screen_list[i].u_row) && (mouser+scroll_baserow < screen_list[i].l_row)
               && (mousec >= screen_list[i].l_col) && (mousec < screen_list[i].r_col))
            menuindex = i;
            /* return LAST menuindex, so we can identify elements in a rung */
         }
      else
         {
         if ((mouser >= screen_list[i].u_row) && (mouser < screen_list[i].l_row)
               && (mousec >= screen_list[i].l_col) && (mousec < screen_list[i].r_col))
            menuindex = i;
            /* return LAST menuindex, so we can identify elements in a rung */
         }
      }
   return(menuindex);
   }

void bad_filesave(savefile,filename)
   FILE savefile;
   BYTE *filename;
   {
   BYTE answer[8];
   BYTE header[64];

   if (savefile != NO_FILE)
      close(savefile);
   unlink(filename);
   sprintf(header,"Cannot save %s!",filename);
   draw_popbox(header);
   cstrout(POPBOX_ROW + 4 * TEXTHEIGHT, POPBOX_COL + 3 * TEXTWIDTH,"Press <enter> to continue",POPBOX_FOREGROUND,POPBOX_BACKGROUND);
   getstr(POPBOX_ROW + 4 * TEXTHEIGHT, POPBOX_COL + (3 + 25) * TEXTWIDTH, answer, 1,POPBOX_FOREGROUND,POPBOX_BACKGROUND);
   show_ladder();
   }

void filesave(menuindex)
   WORD menuindex;
   {
   BYTE filename[16], answer[8];
   WORD i, j;
   FILE savefile;

   null_mouse_object();
   draw_popbox("Save Ladder");
   cstrout(POPBOX_ROW + 4 * TEXTHEIGHT, POPBOX_COL + 3 * TEXTWIDTH,"Filename ? ",POPBOX_FOREGROUND,POPBOX_BACKGROUND);
   getstr(POPBOX_ROW + 4 * TEXTHEIGHT, POPBOX_COL + (3 + 11) * TEXTWIDTH, filename, 12,POPBOX_FOREGROUND,POPBOX_BACKGROUND);
   if (!filename[0])
      {
      show_ladder();
      return;
      }
   for (i = 0; filename[i]; i++)
      if (filename[i] == '.')
         break;
   if ((filename[i] == '\0') && (i <= 8))
      strcat(filename,".LAD");
   if ((savefile = open(filename,0)) != NO_FILE)
      {
      close(savefile);
      draw_popbox(filename);
      cstrout(POPBOX_ROW + 4 * TEXTHEIGHT, POPBOX_COL + 3 * TEXTWIDTH,"Overwrite this file (Y/N) ? ",POPBOX_FOREGROUND,POPBOX_BACKGROUND);
      getstr(POPBOX_ROW + 4 * TEXTHEIGHT, POPBOX_COL + (3 + 28) * TEXTWIDTH, answer, 1,POPBOX_FOREGROUND,POPBOX_BACKGROUND);
      erase_popbox();
      if (toupper(answer[0]) != 'Y')
         {
         show_ladder();
         return;
         }
      }
   if ((savefile = creat(filename)) == NO_FILE)
      return(bad_filesave(savefile,filename));
   if (write(savefile,fileversion,FILEVERSION_LEN) != FILEVERSION_LEN)
      return(bad_filesave(savefile,filename));
   if (write(savefile,rungs,sizeof(rungs)) != sizeof(rungs))
      return(bad_filesave(savefile,filename));
   if (write(savefile,used_rungs,sizeof(used_rungs)) != sizeof(used_rungs))
      return(bad_filesave(savefile,filename));
   if (write(savefile,rung_list,sizeof(rung_list)) != sizeof(rung_list))
      return(bad_filesave(savefile,filename));
   if (write(savefile,&ladder_size,sizeof(ladder_size)) != sizeof(ladder_size))
      return(bad_filesave(savefile,filename));
   if (write(savefile,modbus_dataregs,sizeof(modbus_dataregs)) != sizeof(modbus_dataregs))
      return(bad_filesave(savefile,filename));
   if (close(savefile) == NO_FILE)
      return(bad_filesave(-1,filename));
   show_ladder();
   dirty = FALSE;
   }

void cleanup_dirty()
   {
   BYTE answer[8];

   if (dirty)
      {
      draw_popbox("Ladder has been changed");
      cstrout(POPBOX_ROW + 4 * TEXTHEIGHT, POPBOX_COL + 3 * TEXTWIDTH,"Save this ladder first (Y/N) ? ",POPBOX_FOREGROUND,POPBOX_BACKGROUND);
      getstr(POPBOX_ROW + 4 * TEXTHEIGHT, POPBOX_COL + (3 + 31) * TEXTWIDTH, answer, 1,POPBOX_FOREGROUND,POPBOX_BACKGROUND);
      erase_popbox();
      if (toupper(answer[0]) == 'Y')
         filesave();
      }
   }

void bad_fileload(loadfile,filename)
   FILE loadfile;
   BYTE *filename;
   {
   BYTE answer[8];
   BYTE header[64];

   if (loadfile != NO_FILE)
      close(loadfile);
   sprintf(header,"Cannot load %s!",filename);
   draw_popbox(header);
   cstrout(POPBOX_ROW + 4 * TEXTHEIGHT, POPBOX_COL + 3 * TEXTWIDTH,"Press <enter> to continue",POPBOX_FOREGROUND,POPBOX_BACKGROUND);
   getstr(POPBOX_ROW + 4 * TEXTHEIGHT, POPBOX_COL + (3 + 25) * TEXTWIDTH, answer, 1,POPBOX_FOREGROUND,POPBOX_BACKGROUND);
   newladder();
   }

void fileload(menuindex)
   WORD menuindex;
   {
   BYTE filename[16], answer[8];
   WORD i, j;
   FILE loadfile;

   cleanup_dirty();
   null_mouse_object();
   draw_popbox("Load Ladder");
   cstrout(POPBOX_ROW + 4 * TEXTHEIGHT, POPBOX_COL + 3 * TEXTWIDTH,"Filename ? ",POPBOX_FOREGROUND,POPBOX_BACKGROUND);
   getstr(POPBOX_ROW + 4 * TEXTHEIGHT, POPBOX_COL + (3 + 11) * TEXTWIDTH, filename, 12,POPBOX_FOREGROUND,POPBOX_BACKGROUND);
   erase_popbox();
   if (!filename[0])
      {
      show_ladder();
      return;
      }
   for (i = 0; filename[i]; i++)
      if (filename[i] == '.')
         break;
   if ((filename[i] == '\0') && (i <= 8))
      strcat(filename,".LAD");
   if ((loadfile = open(filename,0)) == NO_FILE)
      {
      close(loadfile);
      draw_popbox(filename);
      cstrout(POPBOX_ROW + 4 * TEXTHEIGHT, POPBOX_COL + 3 * TEXTWIDTH,"This file cannot be found",POPBOX_FOREGROUND,POPBOX_BACKGROUND);
      cstrout(POPBOX_ROW + 6 * TEXTHEIGHT, POPBOX_COL + 3 * TEXTWIDTH,"Press <enter> to continue",POPBOX_FOREGROUND,POPBOX_BACKGROUND);
      getstr(POPBOX_ROW + 6 * TEXTHEIGHT, POPBOX_COL + (3 + 28) * TEXTWIDTH, answer, 1,POPBOX_FOREGROUND,POPBOX_BACKGROUND);
      show_ladder();
      return;
      }
   if (read(loadfile,tempstr,FILEVERSION_LEN) != FILEVERSION_LEN)
      return(bad_filesave(loadfile,filename));
   for (i = 0; i < FILEVERSION_LEN; i++)
      if (tempstr[i] != fileversion[i])
         break;
   if (i != FILEVERSION_LEN)
      {
      close(loadfile);
      draw_popbox(filename);
      cstrout(POPBOX_ROW + 4 * TEXTHEIGHT, POPBOX_COL + 3 * TEXTWIDTH,"Unknown file format",POPBOX_FOREGROUND,POPBOX_BACKGROUND);
      cstrout(POPBOX_ROW + 4 * TEXTHEIGHT, POPBOX_COL + 3 * TEXTWIDTH,"Press <enter> to continue",POPBOX_FOREGROUND,POPBOX_BACKGROUND);
      getstr(POPBOX_ROW + 6 * TEXTHEIGHT, POPBOX_COL + (3 + 28) * TEXTWIDTH, answer, 1,POPBOX_FOREGROUND,POPBOX_BACKGROUND);
      show_ladder();
      return;
      }
   dirty = FALSE;  /* from here on, success or fail, we are "clean" */
   if (read(loadfile,rungs,sizeof(rungs)) != sizeof(rungs))
      return(bad_fileload(loadfile,filename));
   if (read(loadfile,used_rungs,sizeof(used_rungs)) != sizeof(used_rungs))
      return(bad_fileload(loadfile,filename));
   if (read(loadfile,rung_list,sizeof(rung_list)) != sizeof(rung_list))
      return(bad_fileload(loadfile,filename));
   if (read(loadfile,&ladder_size,sizeof(ladder_size)) != sizeof(ladder_size))
      return(bad_fileload(loadfile,filename));
   if (read(loadfile,modbus_dataregs,sizeof(modbus_dataregs)) != sizeof(modbus_dataregs))
      return(bad_fileload(loadfile,filename));
   close(loadfile);
   screen_index = NUM_MENU_ELEMENTS;
   for (i = 0; i < MAX_RUNGS; i++)
      {
      if (used_rungs[i])
         {
         rungs[i].info.left_action = ladder_node_add;
         rungs[i].info.right_action = ladder_node_del;
         screen_list[screen_index++] = (struct element_info *)&rungs[i];
         }
      }
   show_ladder();
   }

void ladder_node_add(menuindex)
   WORD menuindex;
   {
   WORD r, c;
   WORD index;
   WORD rung;

   for (rung = 0; rung < MAX_RUNGS; rung++)
      if ((used_rungs[rung])
             && (mouser+scroll_baserow >= rungs[rung].info.u_row) && (mouser+scroll_baserow < rungs[rung].info.l_row)
             && (mousec >= rungs[rung].info.l_col) && (mousec < rungs[rung].info.r_col))
         break;
   if (rung == MAX_RUNGS)  /* can't find correct rung */
      {
      strout(500,500,"Oops!  can't find rung!  ");
      return;
      }
   if ((index = (mousec - rungs[rung].info.l_col) / LADDER_ELEMENT_WIDTH) >= rungs[rung].num_elements)
      {
      strout(500,500,"1: Oops!  index > ELEMENTS_PER_RUNG  ");
      return;
      }
   if (mouse_object == ORBRANCH_OBJ)
      {
      add_or_branch(rung,index);
      show_ladder();
      return;
      }
   if (mouse_object == EXPAND_OBJ)
      {
      expand_or_branch(rung,index);
      show_ladder();
      return;
      }
#ifndef SIMPLE_ADDROW
   if (mouse_object == ADDROW_OBJ)
      {
      null_mouse_object();
      for (r = 0; r < ladder_size; r++)
         if (rung_list[r] == rung)
            break;
      if (r != ladder_size)  /* if a "root" rung */
         addrow(r+1);  /* add AFTER current rung */
      show_ladder();
      return;
      }
   if (mouse_object == DELROW_OBJ)
      {
      null_mouse_object();
      return;
      }
#endif
   if (rungs[rung].node[index].objindex == NULL_OBJ)
      {
      if (mouse_object == NULL_OBJ)
         return;  /* nothing to drop */
      rungs[rung].node[index].objindex = mouse_object;
      }
   else
      null_mouse_object();
   dirty = TRUE;
   strout(r = rungs[rung].info.u_row-scroll_baserow,c = rungs[rung].info.l_col+TEXTWIDTH+index*LADDER_ELEMENT_WIDTH,rung_element);
   (*objs[rungs[rung].node[index].objindex].drawfn)(r,c,&rungs[rung].node[index],GREEN,TEXTBACKGROUND);
   (*objs[rungs[rung].node[index].objindex].initfn)(&rungs[rung].node[index]);
   show_ladder();
   }

void ladder_node_del(menuindex)
   WORD menuindex;
   {
   WORD r, c;
   WORD index;
   WORD rung;

   for (rung = 0; rung < MAX_RUNGS; rung++)
      if ((used_rungs[rung])
             && (mouser+scroll_baserow >= rungs[rung].info.u_row) && (mouser+scroll_baserow < rungs[rung].info.l_row)
             && (mousec >= rungs[rung].info.l_col) && (mousec < rungs[rung].info.r_col))
         break;
   if (rung == MAX_RUNGS)  /* can't find correct rung */
      {
      strout(500,500,"Oops!  can't find rung!  ");
      null_mouse_object();
      return;
      }
   if ((index = (mousec - rungs[rung].info.l_col) / LADDER_ELEMENT_WIDTH) >= rungs[rung].num_elements)
      {
      strout(500,500,"2: Oops!  index > ELEMENTS_PER_RUNG  ");
      null_mouse_object();
      return;
      }
#ifndef SIMPLE_ADDROW
   if (mouse_object == ADDROW_OBJ)
      {
      null_mouse_object();
      for (r = 0; r < ladder_size; r++)
         if (rung_list[r] == rung)
            break;
      if (r != ladder_size)  /* if a "root" rung */
         addrow(r);  /* add BEFORE current rung */
      show_ladder();
      return;
      }
   if (mouse_object == DELROW_OBJ)
      {
      null_mouse_object();
      for (r = 0; r < ladder_size; r++)
         if (rung_list[r] == rung)
            break;
      if (r != ladder_size)  /* if a "root" rung */
         delrow(r);  /* delete current rung */
      show_ladder();
      return;
      }
#endif
   null_mouse_object();
   if (rungs[rung].node[index].objindex == NULL_OBJ)
      return;
   (*objs[rungs[rung].node[index].objindex].delfn)(rungs[rung].info.u_row-scroll_baserow,rungs[rung].info.l_col+index*LADDER_ELEMENT_WIDTH,&rungs[rung].node[index]);
   rungs[rung].node[index].objindex = NULL_OBJ;
   dirty = TRUE;
   show_ladder();
   }

void clear_elements(rung)
   WORD rung;
   {
   WORD index, i;

   for (index = 0; index < ELEMENTS_PER_RUNG; index++)
      {
      rungs[rung].node[index].objindex = NULL_OBJ;
      rungs[rung].node[index].or_branch = OR_NULL;
      rungs[rung].node[index].or_parent = OR_NULL;
      rungs[rung].node[index].or_parentnode = OR_NULL;
      rungs[rung].node[index].linkstatus = AVAILABLE;
      for (i = 0; i < OPCODE_MAX; i++)
         rungs[rung].node[index].opflags[i] = UNUSED_TYPE;
      }
   rungs[rung].info.left_action = ladder_node_add;
   rungs[rung].info.right_action = ladder_node_del;
   rungs[rung].info.objindex = LADDER_OBJ;
   dirty = TRUE;
   }

void new_rung(rung,screenrow)
   WORD rung,screenrow;
   {
   WORD index;

   if (screen_index == SCREEN_MAX)
      return;
   screen_list[screen_index++] = (struct element_info *)&rungs[rung];
   rungs[rung].info.u_row = screenrow;
   rungs[rung].info.l_row = screenrow + LADDER_ROW_HEIGHT;
   rungs[rung].info.l_col = LADDER_LEFT_COL;
   rungs[rung].info.r_col = LADDER_RIGHT_COL;
   rungs[rung].num_elements = ELEMENTS_PER_RUNG;
   rungs[rung].max_elements = ELEMENTS_PER_RUNG;
   clear_elements(rung);
   used_rungs[rung] = TRUE;
   }

WORD get_rung_maxrow(rung)  /* recursive! */
   WORD rung;
   {
   WORD j, r, maxrow;

   maxrow = rungs[rung].info.l_row;
   for (j = 0; j < rungs[rung].num_elements; j++)
      {
      if (rungs[rung].node[j].or_branch != OR_NULL)
         if ((r = get_rung_maxrow(rungs[rung].node[j].or_branch)) > maxrow)
            maxrow = r;
      }
   return(maxrow);
   }

WORD get_ladder_maxrow()
   {
   return(get_rung_maxrow(rung_list[ladder_size-1]));
   }

WORD find_root_rung(root,rung)  /* recursive! */
   WORD root, rung;
   {
   WORD i;

   if (root == rung)
      return(TRUE);
   for (i = 0; i < rungs[root].num_elements; i++)
      {
      if (rungs[root].node[i].or_branch != OR_NULL)
         if (find_root_rung(rungs[root].node[i].or_branch,rung))
            return(TRUE);
      }
   return(FALSE);
   }

#ifdef BROKEN_ADD_OR_BRANCH
move_rungs_down(rung,row)  /* recursive! */
   {
   WORD i;

   for (i = 0; i < rungs[rung].num_elements; i++)
      if (rungs[rung].node[i].or_branch != OR_NULL)
         move_rungs_down(rungs[rung].node[i].or_branch,row);
   if (rungs[rung].info.l_row >= row)
      {
      rungs[rung].info.u_row += LADDER_ROW_INCREMENT;
      rungs[rung].info.l_row += LADDER_ROW_INCREMENT;
      }
   }

WORD encloses(rung,br,row,lcol,rcol)
   WORD rung, *br, row, lcol, rcol;
   {
   WORD i, branch;

   for (i = 0; i < rungs[rung].num_elements; i++)
      if ((branch = rungs[rung].node[i].or_branch) != OR_NULL)
         if ((rungs[branch].info.l_row >= row)
                   && (rungs[branch].info.l_col < lcol)
                   && (rungs[branch].info.r_col > rcol))
            {
            *br = branch;
            return(rungs[branch].info.l_row);
            }
   return(0);
   }

update_branch_spacing(rung,row,lcol,rcol)  /* recursive! */  /* update branch spacing, if needed */
   WORD rung, row, lcol, rcol;  /* rung is an index into rungs[] */
   {
   WORD i, nextr, parent, branch;

   if (rungs[rung].node[0].or_parent != OR_NULL)
      {
      if (nextr = encloses(rung,&branch,row,lcol,rcol))
         {
         if (nextr <= row)
               /* max current row on parent is too small */
            {
            move_rungs_down(branch,row);
            }
         }
      if ((parent = rungs[rung].node[0].or_parent) != OR_NULL)
         update_branch_spacing(parent,row,lcol,rcol);
      }
   else
      {
      nextr = get_rung_maxrow(rung);
      if (nextr < row)
         nextr = row;
      for (i = 0; i < ladder_size; i++)
         if (rung_list[i] == rung)
            break;
      for (i++; i < ladder_size; i++)
         {
         if (rungs[rung_list[i]].info.l_row <= nextr + LADDER_ROW_INCREMENT)
            break;
         }
      for ( ; i < ladder_size; i++)
         {
         move_rungs_down(rung_list[i],0);
         }
      }
   }

add_or_branch(rung,index)
   WORD rung, index;
   {
   WORD i, j;
   WORD root, maxr;
   WORD nextrung;

   if (rungs[rung].node[index].linkstatus != AVAILABLE)
      return;  /* already connected to an OR branch */
   if (rungs[rung].node[index].or_branch != OR_NULL)
      return;
   for (nextrung = 0; nextrung < MAX_RUNGS; nextrung++)
      if (!used_rungs[nextrung])
         break;
   if (nextrung == MAX_RUNGS)  /* full, can't do anything */
      return;
   if (screen_index == SCREEN_MAX)
      return;
#ifdef NEVERDEF
   if (get_ladder_maxrow() - scroll_baserow > MAX_GRAPHICS_ROW - LADDER_ROW_INCREMENT)
      return;
#endif
   for (j = 0; j < ladder_size; j++)
      {
      root = rung_list[j];
      if (find_root_rung(root,rung))
         break;
      }
   rungs[nextrung].info.u_row = rungs[rung].info.u_row + LADDER_ROW_INCREMENT;
   rungs[nextrung].info.l_row = rungs[rung].info.l_row + LADDER_ROW_INCREMENT;
   rungs[nextrung].info.l_col = rungs[rung].info.l_col + index * LADDER_ELEMENT_WIDTH;
   rungs[nextrung].info.r_col = rungs[nextrung].info.l_col + LADDER_ELEMENT_WIDTH;
   rungs[nextrung].num_elements = 1;
   rungs[nextrung].max_elements = ELEMENTS_PER_RUNG - index;
   clear_elements(nextrung);  /* setup new rung */
   rungs[nextrung].node[0].or_parent = rung;  /* set parent */
   rungs[nextrung].node[0].or_parentnode = index;
   update_branch_spacing(rung,
                         rungs[nextrung].info.l_row,
                         rungs[nextrung].info.l_col,
                         rungs[nextrung].info.r_col);  /* update branch spacing, if needed */
   rungs[rung].node[index].or_branch = nextrung;  /* set rung link */
   rungs[rung].node[index].linkstatus = ENDLINK;  /* default: only 1 element, this must be the end */
      /* DO NOT link rung into ladder until AFTER
       * we scan rung/OR branches for position updates!
       */
#ifdef NEVERDEF
   maxr = get_rung_maxrow(root);  /* get max current row on root */
   if (maxr < rungs[nextrung].info.l_row)  /* if we need more space */
      move_rungs_down(rungs[rung].info.l_row);  /* make space for new branch */
#endif
   screen_list[screen_index++] = (struct element_info *)&rungs[nextrung];
   used_rungs[nextrung] = TRUE;
   }

#else  /* BROKEN_ADD_OR_BRANCH */

move_rungs_down(row)  /* recursive! */
   WORD row;
   {
   WORD rung;

   for (rung = 0; rung < MAX_RUNGS; rung++)
      {
      if ((used_rungs[rung]) && (rungs[rung].info.u_row > row))
          {
          rungs[rung].info.u_row += LADDER_ROW_INCREMENT;
          rungs[rung].info.l_row += LADDER_ROW_INCREMENT;
          }
      }
   }

add_or_branch(rung,index)
   WORD rung, index;
   {
   WORD i, j;
   WORD root, maxr;
   WORD nextrung;

   if (rungs[rung].node[index].linkstatus != AVAILABLE)
      return;  /* already connected to an OR branch */
   if (rungs[rung].node[index].or_branch != OR_NULL)
      return;
   for (nextrung = 0; nextrung < MAX_RUNGS; nextrung++)
      if (!used_rungs[nextrung])
         break;
   if ((nextrung == MAX_RUNGS)  /* full, can't do anything */
              || (ladder_size == MAX_LADDER))
      return;
   if (screen_index == SCREEN_MAX)
      return;
   if (get_ladder_maxrow() - scroll_baserow > MAX_GRAPHICS_ROW - LADDER_ROW_INCREMENT)
      return;
   for (j = 0; j < ladder_size; j++)
      {
      root = rung_list[j];
      if (find_root_rung(root,rung))
         break;
      }
   maxr = get_rung_maxrow(root);  /* get max current row on root */
   rungs[rung].node[index].or_branch = nextrung;  /* set rung link */
   rungs[rung].node[index].linkstatus = ENDLINK;  /* default: only 1 element, this must be the end */
   rungs[nextrung].info.u_row = rungs[rung].info.u_row + LADDER_ROW_INCREMENT;
   rungs[nextrung].info.l_row = rungs[rung].info.l_row + LADDER_ROW_INCREMENT;
   rungs[nextrung].info.l_col = rungs[rung].info.l_col + index * LADDER_ELEMENT_WIDTH;
   rungs[nextrung].info.r_col = rungs[nextrung].info.l_col + LADDER_ELEMENT_WIDTH;
   rungs[nextrung].num_elements = 1;
   rungs[nextrung].max_elements = ELEMENTS_PER_RUNG - index;
   clear_elements(nextrung);  /* setup new rung */
   rungs[nextrung].node[0].or_parent = rung;  /* set parent */
   rungs[nextrung].node[0].or_parentnode = index;
   if (maxr < rungs[nextrung].info.l_row)  /* if we need more space */
      move_rungs_down(rungs[rung].info.l_row);  /* make space for new branch */
   screen_list[screen_index++] = (struct element_info *)&rungs[nextrung];
   used_rungs[nextrung] = TRUE;
   }
#endif  /* BROKEN_ADD_OR_BRANCH */

void expand_or_branch(rung,index)
   WORD rung, index;
   {
   WORD parent, parentnode;
   WORD expandnode;

   if (rungs[rung].num_elements == rungs[rung].max_elements)  /* full, we can't expand */
      return;
   if ((parent = rungs[rung].node[0].or_parent) == OR_NULL)
      return;  /* root rung, can't expand that! */
   parentnode = rungs[rung].node[0].or_parentnode;
   if (parentnode == rungs[parent].max_elements - 1)
      return;  /* at EOL, can't expand */
   expandnode = rungs[rung].node[0].or_parentnode + rungs[rung].num_elements - 1;
   if (expandnode == rungs[parent].num_elements-1)  /* branch runs to end of parent branch */
      return;
   if (rungs[parent].node[expandnode+1].linkstatus != AVAILABLE)
      return;
#ifdef BROKEN_ADD_OR_BRANCH
   rungs[parent].node[expandnode].linkstatus = AVAILABLE;
#else
   rungs[parent].node[expandnode].linkstatus = TAKEN;
#endif
   rungs[parent].node[expandnode+1].linkstatus = ENDLINK;
   rungs[rung].num_elements++;
   rungs[rung].info.r_col += LADDER_ELEMENT_WIDTH;
   }

void newladder(menuindex)
   WORD menuindex;
   {
   WORD rung;

   cleanup_dirty();
   erase_ladder();
   for (screen_index = 0; screen_index < NUM_MENU_ELEMENTS; screen_index++)
      screen_list[screen_index] = &menu[screen_index];
   null_mouse_object();
   ladder_size = 1;
   for (rung = 1; rung < MAX_RUNGS; rung++)
      used_rungs[rung] = FALSE;
   show_menu();
   new_rung(0,FIRST_LADDER_ROW);
   rung_list[0] = 0;   /* new rung list */
   dirty = FALSE;
   scroll_baserow = 0;
   show_ladder();
   }

void xmit_buf(buf,len)
   BYTE *buf;
   WORD len;
   {
   WORD tickcount, t, next;

   tickcount = 0;
   t = get_timer();
   while (len)
      {
      if (ok_to_send())
         {
         send_byte(*buf++);
         len--;
         }
      if (t != (next = get_timer()))
         {
         t = next;
         if (++tickcount == 36)  /* two seconds */
            break;
         }
      }
   }

void get_ack(msg)
   BYTE *msg;
   {
   BYTE answer[8];
   BYTE answerptr;
   WORD tickcount, t, next;

   answerptr = 0;
   tickcount = 0;
   t = get_timer();
   while (tickcount < 9)  /* 1/2 second */
      {
      while (in_queue_head != in_queue_tail)
         {
         if (answerptr < 8)
            answer[answerptr++] = in_queue[in_queue_tail];
         in_queue_tail = (in_queue_tail + 1) & (QUEUE_SIZE-1);  /* it's a power of 2 */
         }
      if (answerptr == 5)  /* size of ACK */
         break;
      if (t != (next = get_timer()))
         {
         t = next;
         tickcount++;
         }
      }
   if ((answerptr != 5)
          || (answer[0] != MSG_BYTE0)
          || (answer[1] != MSG_BYTE1)
          || (answer[2] != MSG_ACK)
          || (answer[3] != 0)
          || (answer[4] != 0))
      {
      draw_popbox(NULL);
      cstrout(POPBOX_ROW + 2 * TEXTHEIGHT, POPBOX_COL + 3 * TEXTWIDTH,msg,POPBOX_FOREGROUND,POPBOX_BACKGROUND);
      cstrout(POPBOX_ROW + 4 * TEXTHEIGHT, POPBOX_COL + 3 * TEXTWIDTH,"Press <enter> to continue",POPBOX_FOREGROUND,POPBOX_BACKGROUND);
      getstr(POPBOX_ROW + 4 * TEXTHEIGHT, POPBOX_COL + (3 + 25) * TEXTWIDTH, answer, 1,POPBOX_FOREGROUND,POPBOX_BACKGROUND);
      show_ladder();
      }
   }

WORD compile_rung(rung,level,find_label_addresses)  /* recursive function! */
   WORD rung, level, find_label_addresses;
   {
   WORD i, error, len, index, branch;
   BYTE objbuf[8];  /* should only need 4 bytes */

   if (level == 0)
      {
      if (program_len < MAX_PROGRAM)
         program[program_len++] = SETLINETRUE;
      else
        return(COMPILE_PROG_TOO_BIG);
      }
   error = COMPILE_OK;
   for (index = 0; (error == COMPILE_OK) && (index < rungs[rung].num_elements); index++)
      {
      if (rungs[rung].node[index].or_branch != OR_NULL)
         {
         branch = rungs[rung].node[index].or_branch;
         if (program_len < MAX_PROGRAM)
            program[program_len++] = PUSHLINETRUE;
         else
            return(COMPILE_PROG_TOO_BIG);
         }
      if (rungs[rung].node[index].objindex == NULL_OBJ)
         {
         error = COMPILE_OK;
         len = 0;  /* blank node, ignore it */
         }
      else
         {
         if ((error = (*objs[rungs[rung].node[index].objindex].compilefn)(&rungs[rung].node[index],&len,objbuf)) != COMPILE_OK)
            {
            if ((error == COMPILE_UNKNOWN_ADDR) && (find_label_addresses))
               error = COMPILE_OK;  /* ignore this error first pass */
            else
               {
               (*objs[rungs[rung].node[index].objindex].drawfn)(rungs[rung].info.u_row-scroll_baserow,rungs[rung].info.l_col+TEXTWIDTH+index*LADDER_ELEMENT_WIDTH,&rungs[rung].node[index],BLACK,RED);
               return(error);
               }
            }
         }
      if ((program_len + len) >= MAX_PROGRAM)
         return(COMPILE_PROG_TOO_BIG);
      for (i = 0; i < len; i++, program_len++)
         {
         program[program_len] = objbuf[i];
         }
      if (rungs[rung].node[index].linkstatus == ENDLINK)
         {
         if (program_len < MAX_PROGRAM)
            program[program_len++] = ORBRANCH;
         else
            return(COMPILE_PROG_TOO_BIG);
         if ((error = compile_rung(branch,level+1,find_label_addresses)) != COMPILE_OK)
            {
            if ((error == COMPILE_UNKNOWN_ADDR) && (find_label_addresses))
               error = COMPILE_OK;  /* ignore this error first pass */
            else
               return(error);
            }
         }  /* if (... == ENDLINK) */
      }  /* for (index = 0; ...) */
   if (level)
      {
      if (program_len < MAX_PROGRAM)
         program[program_len++] = ORPOPLINETRUE;
      else
         return(COMPILE_PROG_TOO_BIG);
      }
   return(COMPILE_OK);
   }

void compile_error(error)
   WORD error;
   {
   BYTE answer[8];

   draw_popbox("Ladder program error:");
   switch (error)
      {
      case COMPILE_UNKNOWN_PARAMS:
         cstrout(POPBOX_ROW + 4 * TEXTHEIGHT, POPBOX_COL + 3 * TEXTWIDTH,"Node parameters not set",POPBOX_FOREGROUND,POPBOX_BACKGROUND);
         break;
      case COMPILE_DUP_ADDR:
         cstrout(POPBOX_ROW + 4 * TEXTHEIGHT, POPBOX_COL + 3 * TEXTWIDTH,"Duplicate label",POPBOX_FOREGROUND,POPBOX_BACKGROUND);
         break;
      case COMPILE_PROG_TOO_BIG:
         sprintf(tempstr,"Program too big %d ",program_len);
         cstrout(POPBOX_ROW + 4 * TEXTHEIGHT, POPBOX_COL + 3 * TEXTWIDTH,tempstr,POPBOX_FOREGROUND,POPBOX_BACKGROUND);
         break;
      }
   cstrout(POPBOX_ROW + 6 * TEXTHEIGHT, POPBOX_COL + 3 * TEXTWIDTH,"Press <enter> to continue",POPBOX_FOREGROUND,POPBOX_BACKGROUND);
   getstr(POPBOX_ROW + 6 * TEXTHEIGHT, POPBOX_COL + (3 + 25) * TEXTWIDTH, answer, 1,POPBOX_FOREGROUND,POPBOX_BACKGROUND);
   show_ladder();
   }

void download(menuindex)
   WORD menuindex;
   {
   WORD i;
   WORD chksum;
   WORD root;
   WORD error;
   BYTE answer[8];

   null_mouse_object();
   for (i = 0; i < NUM_LABELS; i++)
      label_addresses[i] = NULL_LABEL;
      /* Pass 1:  find label addresses */
   program_len = 0;
   error = COMPILE_OK;
   for (root = 0; (error == COMPILE_OK) && (root < ladder_size); root++)
      {
      error = compile_rung(rung_list[root],0,TRUE);  /* TRUE = find label addreses */
      }
   if (error != COMPILE_OK)
      return(compile_error(error));
      /* Pass 2:  generate final code */
   program_len = 0;
   error = COMPILE_OK;
   for (root = 0; (error == COMPILE_OK) && (root < ladder_size); root++)
      {
      error = compile_rung(rung_list[root],0,FALSE);  /* FALSE = label addreses should be known */
      }
   if (error != COMPILE_OK)
      return(compile_error(error));
   for (chksum = 0, i = 0; i < program_len; i++)
      chksum += program[i];
   tempstr[0] = MSG_BYTE0;
   tempstr[1] = MSG_BYTE1;
   tempstr[2] = MSG_PROGRAM;
   tempstr[3] = (program_len + 2) & 0xff;
   tempstr[4] = (program_len + 2) >> 8;
   xmit_buf(tempstr,5);
   xmit_buf(program,program_len);
   tempstr[0] = chksum & 0xff;
   tempstr[1] = chksum >> 8;
   xmit_buf(tempstr,2);
   get_ack("Download failure (comm error?)");
   }

void onestep(menuindex)
   WORD menuindex;
   {
   null_mouse_object();
   beep();
   }

void linestep(menuindex)
   WORD menuindex;
   {
   null_mouse_object();
   beep();
   }

void run(menuindex)
   WORD menuindex;
   {
   BYTE answer[8];

   null_mouse_object();
   tempstr[0] = MSG_BYTE0;
   tempstr[1] = MSG_BYTE1;
   tempstr[2] = MSG_RUN;
   tempstr[3] = 0;
   tempstr[4] = 0;
   xmit_buf(tempstr,5);
   get_ack("RUN failure (comm error?)");
   }

void stop(menuindex)
   WORD menuindex;
   {
   BYTE answer[8];

   null_mouse_object();
   tempstr[0] = MSG_BYTE0;
   tempstr[1] = MSG_BYTE1;
   tempstr[2] = MSG_STOP;
   tempstr[3] = 0;
   tempstr[4] = 0;
   xmit_buf(tempstr,5);
   get_ack("STOP failure (comm error?)");
   }

#ifdef SIMPLE_ADDROW
void addrow(menuindex)
   WORD menuindex;
   {
   WORD maxr;
   WORD nextrung;

   null_mouse_object();
   for (nextrung = 0; nextrung < MAX_RUNGS; nextrung++)
      if (!used_rungs[nextrung])
         break;
   if ((nextrung == MAX_RUNGS)  /* full, can't do anything */
              || (ladder_size == MAX_LADDER))
      return;
   new_rung(nextrung,maxr + LADDER_ROW_INCREMENT);
   rung_list[ladder_size++] = nextrung;
   show_ladder();
   }

void delrow(menuindex)
   WORD menuindex;
   {
   null_mouse_object();
   beep();
   }

#else  /* SIMPLE_ADDROW */

void appendrow(menuindex)
   WORD menuindex;
   {
   WORD maxr;
   WORD nextrung;

   null_mouse_object();
   for (nextrung = 0; nextrung < MAX_RUNGS; nextrung++)
      if (!used_rungs[nextrung])
         break;
   if ((nextrung == MAX_RUNGS)  /* full, can't do anything */
              || (ladder_size == MAX_LADDER))
      return;
   maxr = get_ladder_maxrow();
   new_rung(nextrung,maxr + LADDER_ROW_INCREMENT);
   rung_list[ladder_size++] = nextrung;
   show_ladder();
   }

void addrow_spacing(rung,size)  /* recursive! */
   {
   WORD i;

   for (i = 0; i < rungs[rung].num_elements; i++)
      if (rungs[rung].node[i].or_branch != OR_NULL)
         addrow_spacing(rungs[rung].node[i].or_branch,size);
   rungs[rung].info.u_row += size;
   rungs[rung].info.l_row += size;
   }

void addrow(index)  /* index into rung_list[] */
   WORD index;
   {
   WORD maxr, nextrung, i;

   for (nextrung = 0; nextrung < MAX_RUNGS; nextrung++)
      if (!used_rungs[nextrung])
         break;
   if ((nextrung == MAX_RUNGS)  /* full, can't do anything */
              || (ladder_size == MAX_LADDER))
      return;
   maxr = get_ladder_maxrow() + LADDER_ROW_INCREMENT;
   if (index != ladder_size)  /* if not end-of-list */
      maxr = rungs[rung_list[index]].info.u_row;  /* get row we're going to replace */
   for (i = index; i < ladder_size; i++)
      addrow_spacing(rung_list[i],LADDER_ROW_INCREMENT+LADDER_ROW_HEIGHT);
   for (i = ladder_size; i > index; i--)
      rung_list[i] = rung_list[i-1];  /* move rungs logically "down" to make space */
   rung_list[i] = nextrung;
   new_rung(nextrung,maxr);
   ladder_size++;
   }

void delrow_spacing(rung,size)  /* recursive! */
   {
   WORD i;

   for (i = 0; i < rungs[rung].num_elements; i++)
      if (rungs[rung].node[i].or_branch != OR_NULL)
         delrow_spacing(rungs[rung].node[i].or_branch,size);
   rungs[rung].info.u_row -= size;
   rungs[rung].info.l_row -= size;
   }

void delrow_rungs(rung)  /* recursive! */
   {
   WORD i;

   for (i = 0; i < rungs[rung].num_elements; i++)
      if (rungs[rung].node[i].or_branch != OR_NULL)
         delrow_rungs(rungs[rung].node[i].or_branch);
   used_rungs[rung] = FALSE;
   }

void delrow(index)
   WORD index;
   {
   WORD spacing, i;

   if (ladder_size == 1)
      return;  /* can't completely empty the ladder! */
   spacing = 0;
   if (index < ladder_size - 1)
      {
      spacing = rungs[rung_list[index+1]].info.l_row - rungs[rung_list[index]].info.l_row;
           /* get "moveup" size */
      }
   delrow_rungs(rung_list[index]);
   for (i = index; i < ladder_size - 1; i++)
      {
      rung_list[i] = rung_list[i+1];
      delrow_spacing(rung_list[i],spacing);
      }
   ladder_size--;
   }
#endif  /* SIMPLE_ADDROW */

void pickup(menuindex)
   WORD menuindex;
   {
   null_mouse_object();
   mouse_object = screen_list[menuindex].objindex;
   strcat(mouse_cursor,objs[mouse_object].label);
   strcpy(mouse_overlay," ");
   strcat(mouse_overlay,objs[mouse_object].overlay);
   }

void quit(menuindex)
   WORD menuindex;
   {
   BYTE answer[8];

   if (!quickexit)
      cleanup_dirty();
   exitflag = TRUE;
   }

show_menu()
   {
   WORD i;

   for (i = 0; i < NUM_MENU_ELEMENTS; i++)
      {
      strout(screen_list[i].u_row,screen_list[i].l_col,objs[screen_list[i].objindex].label);
      overlay(screen_list[i].u_row,screen_list[i].l_col,objs[screen_list[i].objindex].overlay);
      }
   }

void show_rung(rung,level)  /* RECURSIVE! */
   WORD rung, level;
   {
   WORD j, r, c;
   WORD i;

   r = rungs[rung].info.u_row;
   if (r < (scroll_baserow + LADDER_ROW_HEIGHT))
      return;
   r -= scroll_baserow;
   if (r > MAX_GRAPHICS_ROW - LADDER_ROW_INCREMENT + LADDER_ROW_HEIGHT)
      return;
   if (level == 0)
      {
      for (i = 0; i < ladder_size; i++)
         if (rung_list[i] == rung)
            break;
      sprintf(tempstr,"%2d",i+1);
      strout(r,LADDER_LABEL_COL,tempstr);
      }
#ifdef NEVERDEF
   if (level)  /* not root */
      strout(r-LADDER_ROW_HEIGHT,rungs[rung].info.l_col,or_vertical);
#endif
   for (j = 0; j < rungs[rung].num_elements; j++)
      {
      c = rungs[rung].info.l_col + j * LADDER_ELEMENT_WIDTH;

           /* start link */
      if (rungs[rung].node[j].or_branch == OR_NULL)
         {
         if ((level == 0) || (j))
            strout(r,c,or_null);
         else
            strout(r,c,or_start);
         }
      else
         {
         if ((!level) || (j != 0))
            strout(r,c,or_link);
         else
            strout(r,c,or_start_t);
         i = rungs[rungs[rung].node[j].or_branch].info.u_row - scroll_baserow - LADDER_ROW_HEIGHT;
         for ( ; i > r; i -= LADDER_ROW_HEIGHT)
            strout(i,c,or_vertical);
         show_rung(rungs[rung].node[j].or_branch,level+1);
         }

        /* node element */
      strout(r,c+TEXTWIDTH,rung_element);
      if (rungs[rung].node[j].objindex != NULL_OBJ)
         (*objs[rungs[rung].node[j].objindex].drawfn)(r,c+TEXTWIDTH,&rungs[rung].node[j],TEXTFOREGROUND,TEXTBACKGROUND);
        /* end link */
      c += LADDER_ELEMENT_WIDTH - TEXTWIDTH;
      if (rungs[rung].node[j].linkstatus == ENDLINK)
         {
         if ((level) && (j == rungs[rung].num_elements - 1))  /* end of OR branch */
            strout(r,c,or_end_t);
         else
            strout(r,c,or_link);
         }
      else
         {
         if ((level) && (j == rungs[rung].num_elements - 1))  /* end of OR branch */
            strout(r,c,or_end);
         else
            {
            strout(r,c,or_null);
#ifdef SHOW_NODES
            if (j != rungs[rung].num_elements - 1)
               {
               setpixel(r+3,c,TEXTFOREGROUND);
               setpixel(r+4,c,TEXTFOREGROUND);
               }
#endif
            }
         }
      }
   if (rungs[rung].node[0].or_parent != OR_NULL)
      {
      j = rungs[rungs[rung].node[0].or_parent].info.u_row - scroll_baserow;
      for (r -= LADDER_ROW_HEIGHT ; r > j; r -= LADDER_ROW_HEIGHT)
         strout(r,c,or_vertical);
      }
   }

erase_ladder()
   {
   erase_box(0,LADDER_LEFT_COL - 3 * TEXTWIDTH,MAX_GRAPHICS_ROW,LADDER_RIGHT_COL - LADDER_LEFT_COL + 8 * TEXTWIDTH);
   }

show_ladder()
   {
   WORD j;
   WORD r;
   WORD maxr;

   erase_ladder();
   maxr = get_ladder_maxrow() - scroll_baserow;
   for (j = 0; j < ladder_size; j++)
      {
      show_rung(rung_list[j],0);
      }
   for (j = 0; j <= maxr; j += LADDER_ROW_HEIGHT)
      {
      strout(j,LADDER_LEFT_COL - TEXTWIDTH,ladder_left);
      strout(j,LADDER_RIGHT_COL,ladder_right);
      }
   show_scrollbar();
   }

show_scrollbar()
   {
   WORD rung, r, maxr;

   cstrout(SCROLL_UPARROW_ROW,SCROLL_LEFTCOL,scroll_uparrow,DKGREY,BLACK);
   cstrout(SCROLL_DOWNARROW_ROW,SCROLL_LEFTCOL,scroll_downarrow,DKGREY,BLACK);
   for (rung = 0; rung < ladder_size; rung++)
      if (rungs[rung_list[rung]].info.u_row == (scroll_baserow + LADDER_ROW_HEIGHT))
         break;
   r = TEXTHEIGHT;
   if (rung)  /* if we can scroll up */
      {
      strout(SCROLL_UPARROW_ROW,SCROLL_LEFTCOL,scroll_uparrow);
      if (ladder_size > 1)
         maxr = SCROLLBAR_SIZE * rung / (ladder_size - 1);
      else
         maxr = SCROLLBAR_SIZE;
      for ( ; r < maxr; r += TEXTHEIGHT)
         strout(r,SCROLL_LEFTCOL,scroll_bar);
      }
   cstrout(r,SCROLL_LEFTCOL,scroll_box,GREEN,TEXTBACKGROUND);
   for (r += TEXTHEIGHT; r <= SCROLLBAR_SIZE; r += TEXTHEIGHT)
      strout(r,SCROLL_LEFTCOL,scroll_bar);
   if (rung != ladder_size - 1)
      strout(SCROLL_DOWNARROW_ROW,SCROLL_LEFTCOL,scroll_downarrow);
   }

void scrollladder(menuindex)
   WORD menuindex;
   {
   WORD rung, r;

   for (rung = 0; rung < ladder_size; rung++)
      if (rungs[rung_list[rung]].info.u_row == (scroll_baserow + LADDER_ROW_HEIGHT))
         break;
   if (ladder_size > 1)
      r = SCROLLBAR_SIZE * rung / (ladder_size - 1) + TEXTHEIGHT;
   else
      r = 2 * TEXTHEIGHT;
   if (mouser <= SCROLL_UPARROW_ROW + TEXTHEIGHT)
      rung = rung ? rung - 1 : 0;
   else
      if (mouser >= SCROLL_DOWNARROW_ROW)
         rung = (rung == ladder_size - 1) ? rung : rung + 1;
      else
         {
         if (mouser <= r)
            {
            if (rung)
               scroll_baserow = rungs[rung_list[rung-1]].info.u_row - LADDER_ROW_HEIGHT;
                  /* we want the "previous" rung to be the last rung on the screen */
            for ( ; rung > 0; rung--)  /* scroll up one "page" */
               if (((scroll_baserow + LADDER_ROW_HEIGHT) - rungs[rung_list[rung]].info.u_row < MAX_GRAPHICS_ROW)
                   && ((scroll_baserow + LADDER_ROW_HEIGHT) - rungs[rung_list[rung-1]].info.u_row >= MAX_GRAPHICS_ROW))
                  break;
            }
         else
            if (rung != ladder_size-1)
               {
               for ( ; rung < ladder_size - 1; rung++)  /* scroll down one "page" */
                  if (rungs[rung_list[rung]].info.u_row - (scroll_baserow + LADDER_ROW_HEIGHT) >= MAX_GRAPHICS_ROW)
                     break;
               }
         }
   scroll_baserow = rungs[rung_list[rung]].info.u_row - LADDER_ROW_HEIGHT;
   show_ladder();
   }

erase_box(r,c,numrows,numcols)
   {
#asm
   push es
   cmp  word [bp+8],0   ;get rows
   jz   skip_erase_box
   cmp  word [bp+10],0  ;and cols
   jz   skip_erase_box
   mov  ax,0a000h
   mov  es,ax
erase_box_loop:
   mov  ax,word [bp+4]   ;get row
   mov  bx,1024  ;bytes per line
   mul  bx
   push ax
   cmp  dx,word videopage_
   je   erase_box_havebank
   mov  word videopage_,dx
   mov  bx,0   ;bh = subfn = select memory window, bl = window = A
   mov  ax,4f05h
   int  10h
erase_box_havebank:
   pop  di
   add  di,word [bp+6]  ;plus cols
   mov  cx,word [bp+10] ;get number of cols
   cld
   xor  ax,ax
   clc
   rcr  cx,1
   jcxz erase_box_skipit
rep stosw
erase_box_skipit:
   jnc  erase_box_skipodd
   stosb
erase_box_skipodd:
   inc  word [bp+4]   ;next row
   dec  word [bp+8]   ;count this wor
   jnz  erase_box_loop
skip_erase_box:
   pop  es
#endasm
   }

gchar()
   {
#asm
   xor  al,al             ;set ZR, clear CY
   mov  ah,11h            ;fn = 101-key check for char
   int  16h               ;call BIOS
   jnae g_a_c_2           ;error, check normal way
   jnz  get_the_char      ;have it, get it
g_a_c_2:
   mov  ah,1              ;check for char
   int  16h               ;call BIOS
   jnz  get_the_char      ;no char, we're done
   xor  ax,ax             ;return is 0
   jmp  get_a_char_done   ;and quit
get_the_char:
   mov  ah,0              ;fn = get char
   int  16h               ;call BIOS
   or   al,al             ;char = 0 (special char?)
   jnz  get_a_char_normal ;no, skip this
   mov  al,ah             ;scancode to al
   mov  ah,1              ;set fnkey flag
   jmp  get_a_char_done   ;and done
get_a_char_normal:
   mov  ah,0              ;clear scancode
;   cmp  al,'a'            ;lower case?
;   jnae get_a_char_done   ;no, skip this
;   cmp  al,'z'            ;still lower case?
;   ja   get_a_char_done   ;no, skip this
;   sub  al,20h            ;force upper case
get_a_char_done:
#endasm
   }

getstr(row,col,str,maxstr,foreground,background)
   WORD row, col;
   char *str;
   WORD maxstr,foreground,background;
   {
   int ch, i;
   char onechar[2];

   while (csts())
      ci();  /* clear keyboard queue */
   onechar[1] = '\0';
   i = 0;
   do
      {
      onechar[0] = '_';
      cstrout(row,col,onechar,RED,background);  /* draw cursor */
      do
         {
         ch = gchar();
         } while (!ch);
      onechar[0] = ' ';
      cstrout(row,col,onechar,foreground,background);  /* erase cursor */
      if ((ch > ' ') && (ch < 0x7f) && (i < maxstr))
         {
         str[i++] =
         onechar[0] = toupper(ch);
         cstrout(row,col,onechar,foreground,background);
         col += TEXTWIDTH;
         }
      if ((ch == '\010') && (i))  /* backspace */
         {
         i--;
         onechar[0] = ' ';
         col -= TEXTWIDTH;
         cstrout(row,col,onechar,foreground,background);
         }
      if (ch == ESC)
         i = 0;  /* return blank string */
      } while ((ch != CR) && (ch != ESC) && (ch != '\t'));
   str[i] = '\0';
   while (csts())
      ci();  /* clear keyboard queue */
   }

draw_popbox(label)
   BYTE *label;
   {
   WORD r, c, i, j, pos;
   BYTE boxline[POPBOX_SIZE + 2];

   for (i = 1; i < POPBOX_SIZE - 1; i++)
      boxline[i] = ' ';
   boxline[i+1] = 0;  /* set EOS */
   boxline[POPBOX_SIZE - 1] =
   boxline[0] = BOX_SIDEBAR;
   for (r = POPBOX_ROW + TEXTHEIGHT, i = 1; i < POPBOX_MAXROWS; i++, r += TEXTHEIGHT)
      cstrout(r,POPBOX_COL,boxline,POPBOX_FOREGROUND,POPBOX_BACKGROUND);
   for (i = 1; i < POPBOX_SIZE - 1; i++)
      boxline[i] = BOX_TOPBAR;
   boxline[0] = BOX_TOP_LEFT;
   boxline[POPBOX_SIZE - 1] = BOX_TOP_RIGHT;
   cstrout(POPBOX_ROW,POPBOX_COL,boxline,POPBOX_FOREGROUND,POPBOX_BACKGROUND);
   boxline[0] = BOX_BOTTOM_LEFT;
   boxline[POPBOX_SIZE - 1] = BOX_BOTTOM_RIGHT;
   cstrout(POPBOX_ROW + TEXTHEIGHT * POPBOX_MAXROWS,POPBOX_COL,boxline,POPBOX_FOREGROUND,POPBOX_BACKGROUND);
   if (label)
      cstrout(POPBOX_ROW + 2 * TEXTHEIGHT, POPBOX_COL + 3 * TEXTWIDTH,label,POPBOX_FOREGROUND,POPBOX_BACKGROUND);
   }

erase_popbox()
   {
   erase_box(POPBOX_ROW,POPBOX_COL,(POPBOX_MAXROWS + 1) * TEXTHEIGHT,POPBOX_SIZE * TEXTWIDTH);
   }

get_values(label,q1,q2,q3,a1,a2,a3)
   BYTE *label, *q1, *q2, *q3, *a1, *a2, *a3;
   {
   WORD pos;

   draw_popbox(label);
   pos = 0;
   if (a1)
      *a1 = 0;
   if (a2)
      *a2 = 0;
   if (a3)
      *a3 = 0;
   if (q1)
      {
      cstrout(POPBOX_ROW + 4 * TEXTHEIGHT, POPBOX_COL + 3 * TEXTWIDTH,q1,POPBOX_FOREGROUND,POPBOX_BACKGROUND);
      if (strlen(q1) > pos)
         pos = strlen(q1);
      }
   if (q2)
      {
      cstrout(POPBOX_ROW + 5 * TEXTHEIGHT, POPBOX_COL + 3 * TEXTWIDTH,q2,POPBOX_FOREGROUND,POPBOX_BACKGROUND);
      if (strlen(q2) > pos)
         pos = strlen(q2);
      }
   if (q3)
      {
      cstrout(POPBOX_ROW + 6 * TEXTHEIGHT, POPBOX_COL + 3 * TEXTWIDTH,q3,POPBOX_FOREGROUND,POPBOX_BACKGROUND);
      if (strlen(q3) > pos)
         pos = strlen(q3);
      }
   if ((pos) && (q1))
      {
      getstr(POPBOX_ROW + 4 * TEXTHEIGHT, POPBOX_COL + (3 + pos) * TEXTWIDTH, a1, 10,POPBOX_FOREGROUND,POPBOX_BACKGROUND);
      if (a1[0] == 0)
         pos = 0;  /* don't ask any more questions */
      }
   if ((pos) && (q2))
      {
      getstr(POPBOX_ROW + 5 * TEXTHEIGHT, POPBOX_COL + (3 + pos) * TEXTWIDTH, a2, 10,POPBOX_FOREGROUND,POPBOX_BACKGROUND);
      if (a2[0] == 0)
         pos = 0;  /* don't ask any more questions */
      }
   if ((pos) && (q3))
      getstr(POPBOX_ROW + 6 * TEXTHEIGHT, POPBOX_COL + (3 + pos) * TEXTWIDTH, a3, 10,POPBOX_FOREGROUND,POPBOX_BACKGROUND);
   erase_popbox();
   }

init_mouse()
   {
#asm
   xor  ax,ax              ;get a 0
   mov  es,ax              ;address INT seg
   mov  ax,word es:[0cch]  ;get INT 33 ofs
   or   ax,word es:[0ceh]  ;OR-in INT 33 seg
   jz   init_mouse_done    ;no INT vector, we're done
   xor  ax,ax              ;fn = check for installed
   int  33h                ;call mouse
   cmp  ax,0ffffh          ;mouse installed?
   mov  ax,0               ;default to NO
   jne  init_mouse_done    ;no mouse, we're done
   mov  ax,0bh             ;fn = read motion counters (set base reference)
   int  33h                ;call mouse
   mov  ax,1               ;set MOUSE to TRUE
init_mouse_done:
#endasm
   }

check_mouse_buttons()
   {
#asm
   xor  si,si            ;clear flags
   mov  ax,5             ;fn = get button press
   xor  bx,bx            ;button = left
   int  33h              ;call mouse
   and  ax,1             ;isolate left button bit
   or   si,ax            ;save button
   mov  ax,5             ;fn = get button press
   mov  bx,1             ;button = right
   int  33h              ;call mouse
   and  ax,2             ;isolate right button bit
   or   si,ax            ;save button
   mov  ax,5             ;fn = get button press
   mov  bx,2             ;button = middle
   int  33h              ;call mouse
   and  ax,4             ;isolate middle button bit
   or   ax,si            ;get other buttons
#endasm
   }

init_mouse_cursor()
   {
   mouser = MAX_GRAPHICS_ROW / 2;
   mousec = MAX_GRAPHICS_COL / 2;
   show_mouse_cursor();
   }

update_mouse_cursor(minr,maxr,minc,maxc)
   int minr,maxr,minc,maxc;
   {
   int r, c;  /* SIGNED, can be + or - */

#asm
   mov  ax,0bh              ;fn = read motion counters
   int  33h                 ;call mouse
   mov  word [bp-2],dx      ;save vertical (row) mickeys
   mov  word [bp-4],cx      ;save horizontal (col) mickeys
#endasm
   if (!( r | c ))  /* no update, do nothing */
      return;
   erase_mouse_cursor();
   mouser += r;  /* update Row */
   if (mouser < minr)
      mouser = minr;
   if (mouser >= maxr)
      mouser = maxr - 1;
   mousec += c;  /* update Col */
   if (mousec < minc)
      mousec = minc;
   if (mousec >= maxc)
      mousec = maxc - 1;
   show_mouse_cursor();
   }

show_mouse_cursor()
   {
   int r, c, i, j;

   if ((mouser == MAX_GRAPHICS_ROW) || (mousec == MAX_GRAPHICS_COL))
      return;
   r = MIN(mouser + MOUSE_SAVE_ROWS,MAX_GRAPHICS_ROW);
   c = MIN(mousec + MOUSE_SAVE_COLS,MAX_GRAPHICS_COL) - mousec;
   for (j = 0, i = mouser; i < r; i++, j++)
      {
      get_pixels(i,mousec,c,mouse_cursor_save[j]);
      }
   overlay(mouser,mousec,mouse_cursor);
   overlay(mouser,mousec,mouse_overlay);
   }

erase_mouse_cursor()
   {
   int r, c, i, j;

   if ((mouser == MAX_GRAPHICS_ROW) || (mousec == MAX_GRAPHICS_COL))
      return;
   r = MIN(mouser + MOUSE_SAVE_ROWS,MAX_GRAPHICS_ROW);
   c = MIN(mousec + MOUSE_SAVE_COLS,MAX_GRAPHICS_COL) - mousec;
   for (j = 0, i = mouser; i < r; i++, j++)
      set_pixels(i,mousec,c,mouse_cursor_save[j]);
   }

null_mouse_object()
   {
   mouse_cursor[0] = MOUSE_CURSOR;
   mouse_cursor[1] = 0;
   mouse_overlay[0] = 0;
   mouse_object = NULL_OBJ;
   }

update_mouse()
   {
   WORD buttons;
   static middle = FALSE;
   static right = FALSE;
   static left = FALSE;
   WORD menuindex;

   update_mouse_cursor(0,MAX_GRAPHICS_ROW,0,MAX_GRAPHICS_COL);
   if ((buttons = check_mouse_buttons()) & (RIGHT_BUTTON | LEFT_BUTTON | MIDDLE_BUTTON))
      menuindex = mouseover_menuindex();
   if (buttons & MIDDLE_BUTTON)
      {
      if (!middle)
         {
         middle = TRUE;
         erase_mouse_cursor();
         mousecolor ^= 0x01;
         show_mouse_cursor();
         }
      }
   else
      if (middle)
         {
         middle = FALSE;
         erase_mouse_cursor();
         mousecolor ^= 0x01;
         show_mouse_cursor();
         }
   if (buttons & RIGHT_BUTTON)
      {
      if (!right)
         {
         right = TRUE;
         erase_mouse_cursor();
         mousecolor ^= 0x02;
         if ((menuindex != NULL_OBJECT)
                && (screen_list[menuindex].right_action != NULL))
            (*screen_list[menuindex].right_action)(menuindex);
         else
            {
            null_mouse_object();
            }
         show_mouse_cursor();
         }
      }
   else
      if (right)
         {
         right = FALSE;
         erase_mouse_cursor();
         mousecolor ^= 0x02;
         show_mouse_cursor();
         }
   if (buttons & LEFT_BUTTON)
      {
      if (!left)
         {
         left = TRUE;
         erase_mouse_cursor();
         mousecolor ^= 0x04;
         if ((menuindex != NULL_OBJECT)
                && (screen_list[menuindex].left_action != NULL))
            (*screen_list[menuindex].left_action)(menuindex);
         show_mouse_cursor();
         }
      }
   else
      if (left)
         {
         left = FALSE;
         erase_mouse_cursor();
         mousecolor ^= 0x04;
         show_mouse_cursor();
         }
   }

clear_inqueue()
   {
   in_queue_tail =
   in_queue_head = 0;  /* clear queue */
   }

BYTE inqueue_next()
   {
   BYTE hold, ch;  /* CH to keep stack aligned */

   hold = in_queue[in_queue_tail];
   in_queue_tail = (in_queue_tail + 1) & (QUEUE_SIZE-1);  /* it's a power of 2 */
   return(hold);
   }

WORD inqueue_size()
   {
   WORD size, tail;

   size = 0;
   tail = in_queue_tail;
   while (tail != in_queue_head)
      {
      size++;
      tail = (tail + 1) & (QUEUE_SIZE-1);  /* it's a power of 2 */
      }
   return(size);
   }

int iabs(val)
   int val;
   {
   if (val < 0)
      return(-val);
   return(val);
   }

add_out_queue(ch)
   BYTE ch;
   {
   out_queue[out_queue_head] = ch;
   out_queue_head = (out_queue_head + 1) & (QUEUE_SIZE - 1);  /* it's a power of 2 */
   }

add_out_queue_word(data)
   WORD data;
   {
   add_out_queue(data);
   add_out_queue(data >> 8);
   }

int ok_to_send()
   {
   if ((inp(commbase+LINE_STATUS) & 0x20) == 0)
      return(FALSE);
   return(TRUE);
   }

send_byte(ch)
   BYTE ch;
   {
   outp(commbase,ch);
   }

int inp(portnum)
   WORD portnum;
   {
#asm
   mov  dx,word [bp+4]   ;get portnum
   in   al,dx            ;get BYTE port
   mov  ah,0             ;make it a WORD
#endasm
   }

outp(portnum,data)
   WORD portnum,data;
   {
#asm
   mov  dx,word [bp+4]       ;get port number
   mov  ax,word [bp+6]       ;get value (BYTE in AL)
   out  dx,al                ;send to I/O port
#endasm
   }

take_comm_interrupt()
   {
#asm
   mov  ax,ds                   ;get ds
   mov  word cs:dssave,ax       ;save DS for interrupt use

   mov  bx,word irqnum_         ;get IRQ number in use
   add  bx,8                    ;make it an INT number
   add  bx,bx                   ;*2
   add  bx,bx                   ;*4 for INT offset
   push ds                      ;save DS
   xor  ax,ax                   ;get a 0
   mov  ds,ax                   ;address INT seg
   mov  si,word [bx]            ;get current int ofs
   mov  di,word [bx+2]          ;get current int seg
   mov  word [bx],offset comm_int  ;set new int ofs
   mov  word [bx+2],cs          ;set new int seg
   pop  ds                      ;restore ds
   mov  word comm_int_ofs_,si   ;save old int ofs
   mov  word comm_int_seg_,di   ;save old int seg
   jmp  take_comm_interrupt_end ;and quit

dssave:  dw  0

comm_int:
   push ax                      ;save regs
   push bx
   push dx
   push si
   push ds
   mov  ds,cs:dssave            ;get local DS
   mov  dx,word commbase_       ;get I/O port base
   in   al,dx                   ;get COMM input char
   mov  ah,0                    ;make it a WORD
   mov  bx,word in_queue_head_  ;get head pointer
   mov  si,offset in_queue_     ;get base
   mov  word [bx+si],ax         ;put into buff
   inc  bx                      ;update pointer
   and  bx,0fffh                ;mod queue size
   mov  word in_queue_head_,bx  ;set new pointer
   mov  dx,20h                  ;8259 I/O port
   mov  ax,20h                  ;EOI signal
   out  dx,al                   ;reset 8259
   pop  ds                      ;restore regs
   pop  si
   pop  dx
   pop  bx
   pop  ax
   iret

take_comm_interrupt_end:
#endasm
   }

reset_comm_interrupt()
   {
   outp(0x21,inp(0x21) | (1 << irqnum));  /* disable COMM interrupts */

#asm
   mov  si,word comm_int_ofs_   ;get saved int ofs
   mov  di,word comm_int_seg_   ;get saved int seg
   mov  bx,word irqnum_         ;get IRQ number in use
   add  bx,8                    ;make it an INT number
   add  bx,bx                   ;*2
   add  bx,bx                   ;*4 for INT offset
   push ds                      ;save DS
   xor  ax,ax                   ;get a 0
   mov  ds,ax                   ;address INT seg
   mov  word [bx],si            ;get current int ofs
   mov  word [bx+2],di          ;get current int seg
   pop  ds                      ;restore ds
#endasm
   }

init_comm_port()
   {
   WORD status, divisor;

   old_uart_setup[INT_EN] = inp(commbase+INT_EN);
   outp(commbase+INT_EN,0);  /* turn off comm interrupts */

   old_uart_setup[LINE_CONTROL] = inp(commbase+LINE_CONTROL);
   old_uart_setup[MODEM_CONTROL] = inp(commbase+MODEM_CONTROL);
   status = inp(commbase+LINE_CONTROL);
   outp(commbase+LINE_CONTROL,status|0x80);
   old_uart_setup[BAUD_LSB] = inp(commbase+BAUD_LSB);
   old_uart_setup[BAUD_MSB] = inp(commbase+BAUD_MSB);

   divisor = FREQUENCY / baudrate;
   outp(commbase+BAUD_LSB,divisor & 0xff);
   outp(commbase+BAUD_MSB,(divisor >> 8) & 0xff);
   status = 0x00;  /* mask: 0x38  0x00 -> parity = NONE */
   status |= 0x03;  /* mask: 0x03  0x03 -> data bits = 8 */
                    /* mask: 0x04  0x00 -> stop bits = 1 */
   outp(commbase+LINE_CONTROL,status);  /* set data format */
   outp(commbase+MODEM_CONTROL,OUT2 | DTR);  /* need to set these */
   outp(commbase+INT_EN,RX_INT);  /* allow interrupts */
   outp(0x21,inp(0x21) & ~(1 << irqnum));  /* enable COMM interrupts */

   }

reset_comm_port()
   {
   outp(commbase+INT_EN,0);  /* disable interrupts */
   outp(commbase+MODEM_CONTROL,old_uart_setup[MODEM_CONTROL]);
   outp(commbase+LINE_CONTROL,inp(commbase+LINE_CONTROL) | 0x80);
   outp(commbase+BAUD_LSB,old_uart_setup[BAUD_LSB]);
   outp(commbase+BAUD_MSB,old_uart_setup[BAUD_MSB]);
   outp(commbase+LINE_CONTROL,old_uart_setup[LINE_CONTROL]);
   }

vesamode(mode)
   int mode;
   {
#asm
   mov  ax,4f02h       ;fn = VESA set video mode
   mov  bx,word [bp+4] ;get video mode
   int  10h            ;call BIOS
#endasm
   videopage = -1;
   }

videomode(mode)
   int mode;
   {
#asm
   mov  ah,0
   mov  al,byte [bp+4]
   int  10h
#endasm
   videopage = -1;
   }

setpixel(r,c,color)
   int r,c,color;
   {
#asm
   push es
   mov  ax,0a000h
   mov  es,ax
   mov  ax,word [bp+4]
   mov  bx,1024  ;bytes per line
   mul  bx
   push ax
   cmp  dx,word videopage_
   je   setpixel_havebank
   mov  word videopage_,dx
   mov  bx,0   ;bh = subfn = select memory window, bl = window = A
   mov  ax,4f05h
   int  10h
setpixel_havebank:
   pop  di
   add  di,word [bp+6]
   mov  al,byte [bp+8]
   stosb
   pop  es
#endasm
   }

getpixel(r,c)
   {
#asm
   push es
   mov  ax,0a000h
   mov  es,ax
   mov  ax,word [bp+4]
   mov  bx,1024  ;bytes per line
   mul  bx
   push ax
   cmp  dx,word videopage_
   je   getpixel_havebank
   mov  word videopage_,dx
   mov  bx,0   ;bh = subfn = select memory window, bl = window = A
   mov  ax,4f05h
   int  10h
getpixel_havebank:
   pop  di
   add  di,word [bp+6]
   xor  ax,ax
   mov  al,byte es:[di]
   pop  es
#endasm
   }

help()
   {
   printf("Oops... that is not right...\n");
   exit(1);
   }

get_vesa_mode()
   {
   int i;
#asm
   mov  word have_vesamode_,0      ;default to no VESA
   mov  word have_vgamode_,0       ;default to no VGA
   mov  word vesalistsize_,0
   mov  ax,4f00h              ;fn = get SuperVGA info
   push ds                    ;data seg
   pop  es                    ;to es
   mov  di,offset buff_512_   ;es:di = 512-byte buffer
   push bp                    ;always save bp
   int  10h                   ;call BIOS
   pop  bp                    ;restore bp
   cmp  ax,004fh              ;AH=0 --> success   AL=4Fh --> function supported
   jne  check_vga             ;nope, how about VGA stuff?
   mov  word have_vesamode_,1      ;set VESA = TRUE
   push ds
   lds  si,word [di+0eh]      ;get ptr to supported VESA modes
   mov  di,offset vesalist_   ;es:di = our mode list
   cld                        ;up!
   xor  cx,cx                 ;count = 0
vesa_check_modes:
   lodsw                      ;get mode
   stosw                      ;save local copy
   inc  cx                    ;count this mode
   inc  ax                    ;was this 0ffffh = end of table?
   jne  vesa_check_modes      ;no, try again
   pop  ds                    ;restore ds
   push ds
   pop  es                    ;set es
   mov  word vesalistsize_,cx
check_vga:
video_done:
#endasm
   vesa_mode = 0;  /* NONE */
   for (i = 0; i < vesalistsize; i++)
      {
      if (vesalist[i] == VESA_1024X768_256COLOR)
         vesa_mode = VESA_1024X768_256COLOR;
      }
   }

/* write what should be a library function */
int strncmpi(s1,s2,n)
   BYTE *s1, *s2;
   WORD n;
   {
   while ((n--) && (*s1) && (*s2))
      {
      if (toupper(*s1) != toupper(*s2))
         return(1);  /* not the same */
      s1++;
      s2++;
      }
   if (n == -1)  /* if end of number of bytes */
      return(0);  /* a match! */
   if ((*s1) || (*s2))  /* if not both at EOS */
      return(1);  /* no match */
   return(0);  /* a match */
   }

/* write what should be a library routine */
WORD strcmpi(str1,str2)
   BYTE *str1, *str2;
   {
   WORD temp, temp2;

   while ((*str1) && (*str2))
      {
      temp = ((*str1 >= 'a') && (*str1 <= 'z')) ? (*str1)-0x20 : *str1;
      temp2 = ((*str2 >= 'a') && (*str2 <= 'z')) ? (*str2)-0x20 : *str2;
      if (temp != temp2)
         return(1);  /* Oops! */
      str1++;
      str2++;
      }
   if (*str1 == *str2)  /* make sure both are at EOS */
      return(0);
   return(1);
   }

void pokew(seg,ofs,val)  /* poke a WORD address */
   WORD seg,ofs,val;
   {
#asm
   push   es
   mov    es,word [bp+4]
   mov    bx,word [bp+6]
   mov    ax,word [bp+8]
   mov    word es:[bx],ax
   pop    es
#endasm
   }

strout(r,c,str)
   WORD r, c;
   char *str;
   {
   WORD i, j, k;
   WORD maxr, maxc;
   BYTE *p;  /* int for TEXTWIDTH > 8, change mask in k */

   maxr = MIN(r + TEXTHEIGHT,MAX_GRAPHICS_ROW);
   while (*str)
      {
      maxc = MIN(c + TEXTWIDTH,MAX_GRAPHICS_COL);
      p = &charset16[(*(str++) & 0x7f) * TEXTHEIGHT * TEXTWIDTH / 8];
      for (i = r; i < maxr; i++, p++)
         for (j = c, k = 0x80; j < maxc; j++, k >>= 1)
            if (*p & k)
               setpixel(i,j,TEXTFOREGROUND);
            else
               setpixel(i,j,TEXTBACKGROUND);
      c += TEXTWIDTH;
      }
   }

cstrout(r,c,str,foreground,background)
   WORD r, c;
   char *str;
   WORD foreground,background;
   {
   WORD i, j, k;
   WORD maxr, maxc;
   BYTE *p;  /* int for TEXTWIDTH > 8, change mask in k */

   maxr = MIN(r + TEXTHEIGHT,MAX_GRAPHICS_ROW);
   while (*str)
      {
      maxc = MIN(c + TEXTWIDTH,MAX_GRAPHICS_COL);
      p = &charset16[(*(str++) & 0x7f) * TEXTHEIGHT * TEXTWIDTH / 8];
      for (i = r; i < maxr; i++, p++)
         for (j = c, k = 0x80; j < maxc; j++, k >>= 1)
            if (*p & k)
               setpixel(i,j,foreground);
            else
               setpixel(i,j,background);
      c += TEXTWIDTH;
      }
   }

overlay(r,c,str)
   WORD r, c;
   char *str;
   {
   WORD i, j, k;
   WORD maxr, maxc;
   BYTE *p;  /* int for TEXTWIDTH > 8, change mask in k */

   maxr = MIN(r + TEXTHEIGHT,MAX_GRAPHICS_ROW);
   while (*str)
      {
      maxc = MIN(c + TEXTWIDTH,MAX_GRAPHICS_COL);
      p = &charset16[(*(str++) & 0x7f) * TEXTHEIGHT * TEXTWIDTH / 8];
      for (i = r; i < maxr; i++, p++)
         for (j = c, k = 0x80; j < maxc; j++, k >>= 1)
            if (*p & k)
               setpixel(i,j,TEXTFOREGROUND);
      c += TEXTWIDTH;
      }
   }

coverlay(r,c,str,foreground)
   WORD r, c;
   char *str;
   WORD foreground;
   {
   WORD i, j, k;
   WORD maxr, maxc;
   BYTE *p;  /* int for TEXTWIDTH > 8, change mask in k */

   maxr = MIN(r + TEXTHEIGHT,MAX_GRAPHICS_ROW);
   while (*str)
      {
      maxc = MIN(c + TEXTWIDTH,MAX_GRAPHICS_COL);
      p = &charset16[(*(str++) & 0x7f) * TEXTHEIGHT * TEXTWIDTH / 8];
      for (i = r; i < maxr; i++, p++)
         for (j = c, k = 0x80; j < maxc; j++, k >>= 1)
            if (*p & k)
               setpixel(i,j,foreground);
      c += TEXTWIDTH;
      }
   }

smallstrout(r,c,str)
   WORD r, c;
   char *str;
   {
   WORD i, j, k;
   WORD maxr, maxc;
   BYTE *p;  /* int for TEXTWIDTH > 8, change mask in k */

   maxr = MIN(r + TEXTHEIGHTSMALL,MAX_GRAPHICS_ROW);
   while (*str)
      {
      maxc = MIN(c + TEXTWIDTH,MAX_GRAPHICS_COL);
      p = &charset8[(*(str++) & 0x7f) * TEXTHEIGHTSMALL * TEXTWIDTH / 8];
      for (i = r; i < maxr; i++, p++)
         for (j = c, k = 0x80; j < maxc; j++, k >>= 1)
            if (*p & k)
               setpixel(i,j,TEXTFOREGROUND);
            else
               setpixel(i,j,TEXTBACKGROUND);
      c += TEXTWIDTH;
      }
   }

smalloverlay(r,c,str)
   WORD r, c;
   char *str;
   {
   WORD i, j, k;
   WORD maxr, maxc;
   BYTE *p;  /* int for TEXTWIDTH > 8, change mask in k */

   maxr = MIN(r + TEXTHEIGHTSMALL,MAX_GRAPHICS_ROW);
   while (*str)
      {
      maxc = MIN(c + TEXTWIDTH,MAX_GRAPHICS_COL);
      p = &charset8[(*(str++) & 0x7f) * TEXTHEIGHTSMALL * TEXTWIDTH / 8];
      for (i = r; i < maxr; i++, p++)
         for (j = c, k = 0x80; j < maxc; j++, k >>= 1)
            if (*p & k)
               setpixel(i,j,TEXTFOREGROUND);
      c += TEXTWIDTH;
      }
   }

get_pixels(r,c,count,buf)
   WORD r, c, count;
   BYTE *buf;
   {
#asm
   push ds
   push es

   push ds
   mov  ax,0a000h
   mov  es,ax
   mov  ax,word [bp+4]   ;get row
   mov  bx,1024  ;bytes per line
   mul  bx
   push ax
   cmp  dx,word videopage_
   je   get_pixels_havebank
   push es
   mov  word videopage_,dx
   mov  bx,0   ;bh = subfn = select memory window, bl = window = A
   mov  ax,4f05h
   int  10h
   pop  es
get_pixels_havebank:
   pop  si
   add  si,word [bp+6]
   mov  di,word [bp+10]
   mov  cx,word [bp+8]
   push es
   pop  ds
   pop  es
   cld
   jcxz skip_get_pixels
   rep  movsb
skip_get_pixels:
   pop  es
   pop  ds
#endasm
   }

set_pixels(r,c,count,buf)
   WORD r, c, count;
   BYTE *buf;
   {
#asm
   push es
   mov  ax,0a000h
   mov  es,ax
   mov  ax,word [bp+4]   ;get row
   mov  bx,1024  ;bytes per line
   mul  bx
   push ax
   cmp  dx,word videopage_
   je   set_pixels_havebank
   mov  word videopage_,dx
   mov  bx,0   ;bh = subfn = select memory window, bl = window = A
   mov  ax,4f05h
   int  10h
set_pixels_havebank:
   pop  di
   add  di,word [bp+6]
   mov  si,word [bp+10]
   mov  cx,word [bp+8]
   cld
   jcxz skip_set_pixels
rep movsb
skip_set_pixels:
   pop  es
#endasm
   }

flood(color)
   int color;
   {
#asm
   push es
   mov  ax,0a000h
   mov  es,ax
   mov  dx,0
flood_loop:
   mov  word videopage_,dx
   mov  bx,0   ;bh = subfn = select memory window, bl = window = A
   mov  ax,4f05h
   int  10h
   mov  al,byte [bp+4]
   mov  ah,al
   cld
   xor  di,di
   mov cx,32768
rep stosw
   inc dx
   cmp dx,12
   jne flood_loop
   pop es
#endasm
   }

WORD peekw(seg,ofs)
  WORD seg, ofs;
  {
#asm
   mov  ax,word [bp+4]   ;get seg
   mov  bx,word [bp+6]   ;get ofs
   push ds
   mov  ds,ax
   mov  ax,word [bx]
   pop  ds
#endasm
  }

WORD get_timer()
   {
   return(peekw(0,0x046c));
   }
