/*
 *  Released under the GNU GPL.  See http://www.gnu.org/licenses/gpl.txt
 *
 *  This program is part of TinyPLC
 *
 *  TinyPLC is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundatation; version 2 of the License.
 *
 *  TinyPLC is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 *  for more details.
 */
/*
 *  My thanks to my employer, Athena Controls (www.athenacontrols.com)
 *  for allowing me to open source this project.
 */
/*
 *  TinyPLC kernel
 *  Processor:   Atmel MEGA1281 or Atmel MEGA2561
 *  Compiler:    Codevision AVR C compiler, 1.24.8e
 *
 *  TinyPLC editor/compiler
 *  Processor:   PC-clone, VESA video (mode 0x105, 1024x768x256)
 *  Compiler:    DeSmet C, version 2.51 (PCC version 1.2)
 *
 *  Revision history:
 *  Programmer        Date       Comments
 *  ----------------  ---------  ---------------------------------------------
 *  William Couture   3/19/2006  Original code
 *  William Couture   9/13/2006  Modify for GPL release
 */

#pragma regalloc-
    /* automatic register allocation OFF for now */
#include <mega1281.h>
#include <ctype.h>

#include "tinyplc.h"

REGISTER BYTE linetrue;
REGISTER BYTE tempvar;
REGISTER BYTE r1;
REGISTER BYTE r2;
REGISTER BYTE runflag;
REGISTER BYTE debugflag;
REGISTER WORD program_ip;
REGISTER BYTE program_valid;
REGISTER WORD program_len;
REGISTER BYTE hardrun;
REGISTER BYTE lasthardrun;
#pragma regalloc+
   /* now that we have our variables allocated to regs, let the compiler use the rest */

BYTE softrun;

BYTE program[MAX_PROGRAM + 2];
  /* +2 for paranoia about R1 and R2 and checksum length */

BYTE message[MAX_MSG_SIZE];
WORD message_ptr;
WORD p_len;
WORD incoming_len;
BYTE message_status;

WORD run_nodecount;  /* for debugging */

WORD program_timer;
BYTE program_abort;

eeprom WORD sacrificial_goat;  /* do not use EEPROM location 0 due to powerup bug on some chips */
eeprom WORD eeprom_program_len;
eeprom BYTE eeprom_program[MAX_PROGRAM];
eeprom WORD eeprom_program_checksum;

struct bytes_t
   {
   BYTE b0;
   BYTE b1;
   BYTE b2;
   BYTE b3;
   };
struct words_t
   {
   WORD w0;
   WORD w1;
   };
struct timercounter_t
   {
   BYTE b0;
   BYTE b1;
   WORD w;
   };

union
   {
   struct bytes_t b;
   struct words_t w;
   struct timercounter_t t;
   DWORD v;
   BYTE bytes[4];
   } dataregs[PLC_MEMORY_LOCS], valdest, val1, val2;
                 /* make valdest, val1, val2 global to avoid passing */


BYTE initstack[STACKSIZE];
BYTE initstackptr;
BYTE resultstack[STACKSIZE];
BYTE resultstackptr;

volatile DWORD tick_counter;

volatile BYTE status_led_pwm;
         BYTE status_led_counter;
volatile WORD seconds_counter;
volatile BYTE seconds_tick;
volatile BYTE newtick;
volatile BYTE xmittick;

#if TOTAL_ANALOGS > 0
   enum
      {
      ADC_SETTLING_1,
      ADC_SETTLING_2,
      ADC_VALID
      };
   BYTE adc_state;
   BYTE adc_channel;
   BYTE adc_tick_counter;
   WORD adc_filter[TOTAL_ANALOGS];
#endif

interrupt [TIM0_COMPA] void heartbeat_isr(void)
   {
   status_led_pwm++;
   tick_counter++;
   newtick = TRUE;
   if (++seconds_counter == HEARTBEAT_FREQUENCY)
      {
      seconds_counter = 0;
      seconds_tick++;
      if (program_timer)
         if (--program_timer == 0)
            program_abort = TRUE;
      }

   if (RUNSTOP_SWITCH == RUNSTOP_RUN)
      hardrun = RUNSTOP_RUN;
   else
      hardrun = RUNSTOP_STOP;
   if (hardrun == RUNSTOP_STOP)
      runflag = FALSE;

   if (xmittick)
      xmittick--;

#if TOTAL_ANALOGS > 0
   if (++adc_tick_counter == ADC_CONVERSION_TIME)  /* ADC should be ready */
      {
      adc_tick_counter = 0;
      if (ADCSRA & ADCSRA_ADSC)  /* DONE bit != 0, not finished */
         {
         /* worry about ADC internal errors later */
         }
      else
         {
         if (adc_state == ADC_VALID)
            {
            adc_filter[adc_channel] -= adc_filter[adc_channel] / ADC_FILTER_STRENGTH;
            adc_filter[adc_channel] += ADCW;
            }
         }
      if (++adc_state > ADC_VALID)
         {
         adc_state = ADC_SETTLING_1;
         if (++adc_channel >= TOTAL_ANALOGS)
            adc_channel = 0;
         ADMUX = ADMUX_VALUE | adc_channel;
         }
      ADCSRA = ADCSRA_GO;  /* start next ADC conversion */
      }
#endif  /* TOTAL_ANALOGS */
   }

BYTE tx_buf[TXBUF_SIZE];
BYTE tx_buf_in;  /* index to next byte to be written into tx_buf[] */
volatile BYTE tx_buf_out; /* index to next byte to be read from tx_buf[] */

BYTE rx_buf[RXBUF_SIZE];
BYTE rx_buf_in;  /* index to next byte to be written into rx_buf[] */
BYTE rx_buf_out; /* index to next byte to be read from rx_buf[] */
BYTE rx_byte, rx_temp;
static BYTE stopmsg[5] = { MSG_BYTE0, MSG_BYTE1, MSG_STOP, 0, 0 };

interrupt [USART0_RXC] void usart0_receive(void)
   {
   BYTE i, j, k;

   rx_temp = UCSR0A;
   rx_byte = UDR0;
   if (!(rx_temp & FRAMING_ERROR))
      {
      rx_buf[rx_buf_in] = rx_byte;
      if (rx_byte == 0)
         {
         for (i = 0, j = 4, k = rx_buf_in; i < 5; i++, j--)
            {
            if (rx_buf[k] != stopmsg[j])
               break;
            k = (k == 0) ? RXBUF_SIZE - 1 : k - 1;
            }
         if (i == 5)
            {
            runflag = FALSE;
            softrun = FALSE;
            }
         }
      rx_temp = (rx_buf_in == RXBUF_SIZE - 1) ? 0 : rx_buf_in + 1;
      if (rx_temp != rx_buf_out)
         rx_buf_in = rx_temp;
      }
   }

interrupt [USART0_TXC] void usart0_tx_complete(void)
   {
   UCSR0B = UCSR0B_INIT;  /* turn off TX, stuff in buffer will finish xmiting */
   }

interrupt [USART0_UDRE] void usart0_datareg_empty(void)
   {
   if (tx_buf_out == tx_buf_in)  /* buffer empty */
      {
      UCSR0B = UCSR0B_FINISH;  /* turn off UDRIE0, stuff in buffer will finish xmiting */
      }
   else
      {
      UDR0 = tx_buf[tx_buf_out];
      tx_buf_out = (tx_buf_out == TXBUF_SIZE - 1) ? 0 : tx_buf_out + 1;
      }
   }

void xmit_char(BYTE ch)
   {
   BYTE temp;

   DI();
   UCSR0B = UCSR0B_XMIT;  /* enable transmitter */
   temp = (tx_buf_in == TXBUF_SIZE - 1) ? 0 : tx_buf_in + 1;
   if (temp != tx_buf_out)  /* if buffer not full */
      {
      tx_buf[tx_buf_in] = ch;
      tx_buf_in = temp;
      }
   EI();
   }

void xmit_wait(BYTE ch)
   {
   BYTE temp;

   temp = (tx_buf_in == TXBUF_SIZE - 1) ? 0 : tx_buf_in + 1;
   xmittick = 2;
   while ((temp == tx_buf_out) && (xmittick))  /* if buffer not full */
      ;
   if (xmittick)  /* if not timeout */
      xmit_char(ch);
   }

void xmit_str(char flash *s)
   {
   while (*s)
      xmit_char(*s++);
   }

void system_init(void)
   {
   PET_WATCHDOG();

   WDTCSR = WDTCSR_UNLOCK;   /* unlock WDTCSR register to update Watchdog params */
   WDTCSR = WDTCSR_SET_WDT_PERIOD;  /* set WDT period to 2 seconds */

   DDRA = PORTA_DIRECTIONS;
   PORTA = PORTA_DEFAULT;

   DDRB = PORTB_DIRECTIONS;
   PORTB = PORTB_DEFAULT;

   DDRC = PORTC_DIRECTIONS;
   PORTC = PORTC_DEFAULT;

   DDRD = PORTD_DIRECTIONS;
   PORTD = PORTD_DEFAULT;

   DDRE = PORTE_DIRECTIONS;
   PORTE = PORTE_DEFAULT;

   DDRF = PORTF_DIRECTIONS;
   PORTF = PORTF_DEFAULT;

   DDRG = PORTG_DIRECTIONS;
   PORTG = PORTG_DEFAULT;

   TCCR0A = TCCR0A_INIT;  /* init TMR0 for "heartbeat" */
   TIMSK0 = TIMSK0_INIT;
   OCR0A = TMR0_PERIOD;
   TCCR0B = TCCR0B_INIT;

   UCSR0C = UCSR0C_INIT;  /* init USART 0 */
   UCSR0B = UCSR0B_INIT;
   UBRR0H = UBRR_VALUE >> 8;
   UBRR0L = (UBRR_VALUE & 0xff);

#if TOTAL_ANALOGS > 0
   ADCSRA = ADCSRA_INIT;  /* init ADC */
   DIDR0 = DIDR0_VALUE;   /* set pins for analog input */
   ADMUX = ADMUX_VALUE | adc_channel;
   ADCSRA = ADCSRA_GO;    /* start first ADC conversion */
   adc_tick_counter = 0;
#endif

   EI();

   program_timer = 0;    /* get ready for incoming message */
   program_abort = FALSE;
   message_ptr = 0;
   message_status = MSG_STATUS_WAITING;
   incoming_len = 0xffff;  /* invalid! */
   }


/* set the real outputs */
void service_outputs(void)
   {
      /* NOTE: this could be done as a for() loop, such as
       *            for (i = 0; i < TOTAL_OUTPUTS; i++)
       *               {
       *               val = dataregs[FIRST_OUTPUT + i].v ? OUTPUT_ON : OUTPUT_OFF;
       *               switch (i)
       *                  {
       *         #if TOTAL_OUTPUTS > 0
       *                  case 0:
       *                     DIG_OUT_0  = val;
       *                     break;
       *         #endif
       *                  { etc }
       *                  }
       *               }
       *       but the CASE statement would still need be setup for the
       *       number of outputs.
       */
#if TOTAL_OUTPUTS > 0
   DIG_OUT_0  = dataregs[FIRST_OUTPUT +  0].v ? OUTPUT_ON : OUTPUT_OFF;
#endif
#if TOTAL_OUTPUTS > 1
   DIG_OUT_1  = dataregs[FIRST_OUTPUT +  1].v ? OUTPUT_ON : OUTPUT_OFF;
#endif
#if TOTAL_OUTPUTS > 2
   DIG_OUT_2  = dataregs[FIRST_OUTPUT +  2].v ? OUTPUT_ON : OUTPUT_OFF;
#endif
#if TOTAL_OUTPUTS > 3
   DIG_OUT_3  = dataregs[FIRST_OUTPUT +  3].v ? OUTPUT_ON : OUTPUT_OFF;
#endif
#if TOTAL_OUTPUTS > 4
   DIG_OUT_4  = dataregs[FIRST_OUTPUT +  4].v ? OUTPUT_ON : OUTPUT_OFF;
#endif
#if TOTAL_OUTPUTS > 5
   DIG_OUT_5  = dataregs[FIRST_OUTPUT +  5].v ? OUTPUT_ON : OUTPUT_OFF;
#endif
#if TOTAL_OUTPUTS > 6
   DIG_OUT_6  = dataregs[FIRST_OUTPUT +  6].v ? OUTPUT_ON : OUTPUT_OFF;
#endif
#if TOTAL_OUTPUTS > 7
   DIG_OUT_7  = dataregs[FIRST_OUTPUT +  7].v ? OUTPUT_ON : OUTPUT_OFF;
#endif
#if TOTAL_OUTPUTS > 8
   DIG_OUT_8  = dataregs[FIRST_OUTPUT +  8].v ? OUTPUT_ON : OUTPUT_OFF;
#endif
#if TOTAL_OUTPUTS > 9
   DIG_OUT_9  = dataregs[FIRST_OUTPUT +  9].v ? OUTPUT_ON : OUTPUT_OFF;
#endif
#if TOTAL_OUTPUTS > 10
   DIG_OUT_10 = dataregs[FIRST_OUTPUT + 10].v ? OUTPUT_ON : OUTPUT_OFF;
#endif
#if TOTAL_OUTPUTS > 11
   DIG_OUT_11 = dataregs[FIRST_OUTPUT + 11].v ? OUTPUT_ON : OUTPUT_OFF;
#endif
#if TOTAL_OUTPUTS > 12
   DIG_OUT_12 = dataregs[FIRST_OUTPUT + 12].v ? OUTPUT_ON : OUTPUT_OFF;
#endif
#if TOTAL_OUTPUTS > 13
   DIG_OUT_13 = dataregs[FIRST_OUTPUT + 13].v ? OUTPUT_ON : OUTPUT_OFF;
#endif
#if TOTAL_OUTPUTS > 14
   DIG_OUT_14 = dataregs[FIRST_OUTPUT + 14].v ? OUTPUT_ON : OUTPUT_OFF;
#endif
#if TOTAL_OUTPUTS > 15
   DIG_OUT_15 = dataregs[FIRST_OUTPUT + 15].v ? OUTPUT_ON : OUTPUT_OFF;
#endif
#if TOTAL_OUTPUTS > 16
   DIG_OUT_16 = dataregs[FIRST_OUTPUT + 16].v ? OUTPUT_ON : OUTPUT_OFF;
#endif
#if TOTAL_OUTPUTS > 17
   DIG_OUT_17 = dataregs[FIRST_OUTPUT + 17].v ? OUTPUT_ON : OUTPUT_OFF;
#endif
#if TOTAL_OUTPUTS > 18
   DIG_OUT_18 = dataregs[FIRST_OUTPUT + 18].v ? OUTPUT_ON : OUTPUT_OFF;
#endif
#if TOTAL_OUTPUTS > 19
   DIG_OUT_19 = dataregs[FIRST_OUTPUT + 19].v ? OUTPUT_ON : OUTPUT_OFF;
#endif
#if TOTAL_OUTPUTS > 20
   DIG_OUT_20 = dataregs[FIRST_OUTPUT + 20].v ? OUTPUT_ON : OUTPUT_OFF;
#endif
#if TOTAL_OUTPUTS > 21
   #error This program is only setup to handle up to 21 outputs
#endif
   }

void gather_inputs()
   {
   BYTE i;

      /* NOTE: this could be done as a for() loop, such as
       *            for (i = 0; i < TOTAL_INPUTS; i++)
       *               {
       *               switch (i)
       *                  {
       *         #if TOTAL_INPUTS > 0
       *                  case 0:
       *                     val = (DIG_IN_0  == INPUT_ON) ? ON : OFF;
       *                     break;
       *         #endif
       *                  { etc }
       *                  }
       *               dataregs[FIRST_INPUT + i] = val;
       *               }
       *       but the CASE statement would still need be setup for the
       *       number of inputs.
       */
#if TOTAL_INPUTS > 0
   dataregs[FIRST_INPUT +  0].v = (DIG_IN_0  == INPUT_ON) ? ON : OFF;
#endif
#if TOTAL_INPUTS > 1
   dataregs[FIRST_INPUT +  1].v = (DIG_IN_1  == INPUT_ON) ? ON : OFF;
#endif
#if TOTAL_INPUTS > 2
   dataregs[FIRST_INPUT +  2].v = (DIG_IN_2  == INPUT_ON) ? ON : OFF;
#endif
#if TOTAL_INPUTS > 3
   dataregs[FIRST_INPUT +  3].v = (DIG_IN_3  == INPUT_ON) ? ON : OFF;
#endif
#if TOTAL_INPUTS > 4
   dataregs[FIRST_INPUT +  4].v = (DIG_IN_4  == INPUT_ON) ? ON : OFF;
#endif
#if TOTAL_INPUTS > 5
   dataregs[FIRST_INPUT +  5].v = (DIG_IN_5  == INPUT_ON) ? ON : OFF;
#endif
#if TOTAL_INPUTS > 6
   dataregs[FIRST_INPUT +  6].v = (DIG_IN_6  == INPUT_ON) ? ON : OFF;
#endif
#if TOTAL_INPUTS > 7
   dataregs[FIRST_INPUT +  7].v = (DIG_IN_7  == INPUT_ON) ? ON : OFF;
#endif
#if TOTAL_INPUTS > 8
   dataregs[FIRST_INPUT +  8].v = (DIG_IN_8  == INPUT_ON) ? ON : OFF;
#endif
#if TOTAL_INPUTS > 9
   dataregs[FIRST_INPUT +  9].v = (DIG_IN_9  == INPUT_ON) ? ON : OFF;
#endif
#if TOTAL_INPUTS > 10
   dataregs[FIRST_INPUT + 10].v = (DIG_IN_10 == INPUT_ON) ? ON : OFF;
#endif
#if TOTAL_INPUTS > 11
   dataregs[FIRST_INPUT + 11].v = (DIG_IN_11 == INPUT_ON) ? ON : OFF;
#endif
#if TOTAL_INPUTS > 12
   dataregs[FIRST_INPUT + 12].v = (DIG_IN_12 == INPUT_ON) ? ON : OFF;
#endif
#if TOTAL_INPUTS > 13
   dataregs[FIRST_INPUT + 13].v = (DIG_IN_13 == INPUT_ON) ? ON : OFF;
#endif
#if TOTAL_INPUTS > 14
   dataregs[FIRST_INPUT + 14].v = (DIG_IN_14 == INPUT_ON) ? ON : OFF;
#endif
#if TOTAL_INPUTS > 15
   dataregs[FIRST_INPUT + 15].v = (DIG_IN_15 == INPUT_ON) ? ON : OFF;
#endif
#if TOTAL_INPUTS > 16
   dataregs[FIRST_INPUT + 16].v = (DIG_IN_16 == INPUT_ON) ? ON : OFF;
#endif
#if TOTAL_INPUTS > 17
   dataregs[FIRST_INPUT + 17].v = (DIG_IN_17 == INPUT_ON) ? ON : OFF;
#endif
#if TOTAL_INPUTS > 18
   dataregs[FIRST_INPUT + 18].v = (DIG_IN_18 == INPUT_ON) ? ON : OFF;
#endif
#if TOTAL_INPUTS > 18
   dataregs[FIRST_INPUT + 19].v = (DIG_IN_19 == INPUT_ON) ? ON : OFF;
#endif
#if TOTAL_INPUTS > 20
   dataregs[FIRST_INPUT + 20].v = (DIG_IN_20 == INPUT_ON) ? ON : OFF;
#endif
#if TOTAL_INPUTS > 21
   #error This program is only setup to handle up to 21 inputs
#endif
#if TOTAL_ANALOGS > 0
   for (i = 0; i < TOTAL_ANALOGS; i++)
      dataregs[FIRST_ANALOG + i].v = (adc_filter[i] + ADC_FILTER_ROUNDING) / ADC_FILTER_STRENGTH;
#endif
   }

/* initialize all logic variables */
void init_logic(void)
   {
   WORD i;

   program_ip = 0;

   for (i = 0; i < PLC_MEMORY_LOCS; i++)
      dataregs[i].v = 0;

   service_outputs();  /* start with outputs OFF */
   }

void nop()
   {
   program_ip++;
   }

void setlinetrue()
   {
   program_ip++;
   linetrue = TRUE;
   }

void pushlinetrue()
   {
   program_ip++;
   if (initstackptr == STACKSIZE)
      {
      program_valid = FALSE;
      runflag = FALSE;
      program_ip = program_len;  /* stop program immediately */
      dataregs[STATUSREG].b.b0 |= STATUS_STK;
      linetrue = FALSE;
      }
   initstack[initstackptr++] = linetrue;
   }

void orbranch()
   {
   program_ip++;
   if ((resultstackptr == STACKSIZE) || (initstackptr == 0))
      {
      program_valid = FALSE;
      runflag = FALSE;
      program_ip = program_len;  /* stop program immediately */
      dataregs[STATUSREG].b.b0 |= STATUS_STK;
      linetrue = FALSE;
      }
   resultstack[resultstackptr++] = linetrue;
   linetrue = initstack[initstackptr - 1];
   }

void orpoplinetrue()
   {
   program_ip++;
   if ((initstackptr == 0) || (resultstackptr == 0))
      {
      program_valid = FALSE;
      runflag = FALSE;
      program_ip = program_len;  /* stop program immediately */
      dataregs[STATUSREG].b.b0 |= STATUS_STK;
      linetrue = FALSE;
      }
   linetrue |= resultstack[--resultstackptr];
   initstackptr--;
   }

void inp()
   {
   if (linetrue)
      linetrue = dataregs[r1].v ? TRUE : FALSE;
   program_ip += 2;
   }

void ninp()
   {
   if (linetrue)
      linetrue = dataregs[r1].v ? FALSE : TRUE;
   program_ip += 2;
   }

void out()
   {
   dataregs[r1].v = linetrue;
   program_ip += 2;
   }

void nout()
   {
   dataregs[r1].v = !linetrue;
   program_ip += 2;
   }

void lout()
   {
   if (linetrue)
      dataregs[r1].v = TRUE;
   program_ip += 2;
   }

void uout()
   {
   if (linetrue)
      dataregs[r1].v = FALSE;
   program_ip += 2;
   }

void setmathflags()  /* set status flags for math op */
   {
   dataregs[STATUSREG].b.b0 &= ~(STATUS_CY | STATUS_ZR | STATUS_OV | STATUS_NG);
   if (!valdest.v)
      dataregs[STATUSREG].b.b0 |= STATUS_ZR;
   if (valdest.b.b3 & SIGNBIT)
      {
      dataregs[STATUSREG].b.b0 |= STATUS_NG;
      if ((!(val1.b.b3 & SIGNBIT)) && (!(val2.b.b3 & SIGNBIT)))
         dataregs[STATUSREG].b.b0 |= STATUS_OV;
      }
   if ((valdest.v < val1.v) && (valdest.v < val2.v))
      dataregs[STATUSREG].b.b0 |= STATUS_CY;
   }

void add()
   {
   if (linetrue)
      {
      valdest.v =
      dataregs[program[program_ip+3]].v =
             (val1.v = dataregs[r1].v)
           + (val2.v = dataregs[r2].v);
      setmathflags();
      }
   program_ip += 4;
   }

void addb()
   {
   if (linetrue)
      {
      valdest.v =
      dataregs[program[program_ip+3]].v =
             (val1.v = dataregs[r1].v)
           + (val2.v = (DWORD)r2);
      setmathflags();
      }
   program_ip += 4;
   }

void sub()
   {
   if (linetrue)
      {
      valdest.v =
      dataregs[program[program_ip+3]].v =
             (val1.v = dataregs[r1].v)
           - (val2.v = dataregs[r2].v);
      setmathflags();
      }
   program_ip += 4;
   }

void subb()
   {
   if (linetrue)
      {
      valdest.v =
      dataregs[program[program_ip+3]].v =
             (val1.v = dataregs[r1].v)
           - (val2.v = (DWORD)r2);
      setmathflags();
      }
   program_ip += 4;
   }

void mul()
   {
   if (linetrue)
      {
      valdest.v =
      dataregs[program[program_ip+3]].v =
             (val1.v = dataregs[r1].v)
           * (val2.v = dataregs[r2].v);
      setmathflags();
      }
   program_ip += 4;
   }

void mulb()
   {
   if (linetrue)
      {
      valdest.v =
      dataregs[program[program_ip+3]].v =
             (val1.v = dataregs[r1].v)
           * (val2.v = (DWORD)r2);
      setmathflags();
      }
   program_ip += 4;
   }

void div()
   {
   if (linetrue)
      {
      if ((val2.v = dataregs[r2].v) == 0)
         {
         dataregs[STATUSREG].b.b0 = STATUS_DIV;
         linetrue = FALSE;
         }
      else
         {
         valdest.v =
         dataregs[program[program_ip+3]].v =
                (val1.v = dataregs[r1].v)
              / val2.v;
         setmathflags();
         }
      }
   program_ip += 4;
   }

void divb()
   {
   if (linetrue)
      {
      if ((val2.v = (DWORD)r2) == 0)
         {
         dataregs[STATUSREG].b.b0 = STATUS_DIV;
         linetrue = FALSE;
         }
      else
         {
         valdest.v =
         dataregs[program[program_ip+3]].v =
                (val1.v = dataregs[r1].v)
              / val2.v;
         setmathflags();
         }
      }
   program_ip += 4;
   }

void mod()
   {
   if (linetrue)
      {
      if ((val2.v = dataregs[r2].v) == 0)
         {
         dataregs[STATUSREG].b.b0 = STATUS_DIV;
         linetrue = FALSE;
         }
      else
         {
         valdest.v =
         dataregs[program[program_ip+3]].v =
                (val1.v = dataregs[r1].v)
              % val2.v;
         setmathflags();
         }
      }
   program_ip += 4;
   }

void modb()
   {
   if (linetrue)
      {
      if ((val2.v = (DWORD)r2) == 0)
         {
         dataregs[STATUSREG].b.b0 = STATUS_DIV;
         linetrue = FALSE;
         }
      else
         {
         valdest.v =
         dataregs[program[program_ip+3]].v =
                (val1.v = dataregs[r1].v)
              % val2.v;
         setmathflags();
         }
      }
   program_ip += 4;
   }

void setlogicflags()  /* set status flags for logic op */
   {
   dataregs[STATUSREG].b.b0 &= ~(STATUS_CY | STATUS_ZR | STATUS_OV | STATUS_NG);
   if (!valdest.v)
      dataregs[STATUSREG].b.b0 |= STATUS_ZR;
   }

void and(void linetrue)
   {
   if (linetrue)
      {
      valdest.v =
      dataregs[program[program_ip+3]].v =
             (val1.v = dataregs[r1].v)
           & (val2.v = dataregs[r2].v);
      setlogicflags();
      }
   program_ip += 4;
   }

void andb()
   {
   if (linetrue)
      {
      valdest.v =
      dataregs[program[program_ip+3]].v =
             (val1.v = dataregs[r1].v)
           & (val2.v = (DWORD)r2);
      setlogicflags();
      }
   program_ip += 4;
   }

void or()
   {
   if (linetrue)
      {
      valdest.v =
      dataregs[program[program_ip+3]].v =
             (val1.v = dataregs[r1].v)
           | (val2.v = dataregs[r2].v);
      setlogicflags();
      }
   program_ip += 4;
   }

void orb()
   {
   if (linetrue)
      {
      valdest.v =
      dataregs[program[program_ip+3]].v =
             (val1.v = dataregs[r1].v)
           | (val2.v = (DWORD)r2);
      setlogicflags();
      }
   program_ip += 4;
   }

void xor()
   {
   if (linetrue)
      {
      valdest.v =
      dataregs[program[program_ip+3]].v =
             (val1.v = dataregs[r1].v)
           ^ (val2.v = dataregs[r2].v);
      setlogicflags();
      }
   program_ip += 4;
   }

void xorb()
   {
   if (linetrue)
      {
      valdest.v =
      dataregs[program[program_ip+3]].v =
             (val1.v = dataregs[r1].v)
           ^ (val2.v = (DWORD)r2);
      setlogicflags();
      }
   program_ip += 4;
   }

void not()
   {
   if (linetrue)
      {
      val2.v = 0;
      val1.v =
      valdest.v =
      dataregs[r1].v = ~dataregs[r1].v;
      setlogicflags();
      }
   program_ip += 2;
   }

void neg()
   {
   if (linetrue)
      {
      val2.v = 0;
      val1.v =
      valdest.v =
      dataregs[r1].v = -dataregs[r1].v;
      setmathflags();
      }
   program_ip += 2;
   }

void equ()
   {
   linetrue = (dataregs[r1].v == dataregs[r2].v);
   program_ip += 3;
   }

void equb()
   {
   linetrue = (dataregs[r1].v == (DWORD)r2);
   program_ip += 3;
   }

void neq()
   {
   linetrue = (dataregs[r1].v != dataregs[r2].v);
   program_ip += 3;
   }

void neqb()
   {
   linetrue = (dataregs[r1].v != (DWORD)r2);
   program_ip += 3;
   }

void grt()
   {
   linetrue = (dataregs[r1].v > dataregs[r2].v);
   program_ip += 3;
   }

void grtb()
   {
   linetrue = (dataregs[r1].v > (DWORD)r2);
   program_ip += 3;
   }

void geq()
   {
   linetrue = (dataregs[r1].v >= dataregs[r2].v);
   program_ip += 3;
   }

void geqb()
   {
   linetrue = (dataregs[r1].v >= (DWORD)r2);
   program_ip += 3;
   }

void les()
   {
   linetrue = (dataregs[r1].v < dataregs[r2].v);
   program_ip += 3;
   }

void lesb()
   {
   linetrue = (dataregs[r1].v < (DWORD)r2);
   program_ip += 3;
   }

void leq()
   {
   linetrue = (dataregs[r1].v <= dataregs[r2].v);
   program_ip += 3;
   }

void leqb()
   {
   linetrue = (dataregs[r1].v <= (DWORD)r2);
   program_ip += 3;
   }

void mov()
   {
   if (linetrue)
      {
      dataregs[r1].v = dataregs[r2].v;
      }
   program_ip += 3;
   }

void movd()
   {
   if (linetrue)
      {
      dataregs[r1].v = *(DWORD *)&program[program_ip+2];
      }
   program_ip += 6;
   }

void jmp()
   {
   if (linetrue)
      {
      program_ip = *(WORD *)&program[program_ip+1];
      }
   else
      program_ip += 3;
   }

void osr()
   {
   tempvar = linetrue;  /* need to save this before we change it */
   if (linetrue)
      {
      if (dataregs[r1].b.b0)
         linetrue = FALSE;
      }
   dataregs[r1].b.b0 = tempvar;
   program_ip += 2;
   }

void ctt()
   {
   tempvar = linetrue;  /* need to save this before we change it */
   if (linetrue)
      {
      if (!(dataregs[r1].t.b0 & TIMER_DONE_BIT))  /* if not already done */
         {
         dataregs[r1].t.b0 = TIMER_ENABLED_BIT | TIMER_ACTIVE_BIT;
         if (!dataregs[r1].t.b1)  /* FALSE -> TRUE transition */
            {
            if (++dataregs[r1].t.w >= *(WORD *)&program[program_ip+2])
               dataregs[r1].t.b0 = TIMER_ENABLED_BIT | TIMER_DONE_BIT;
            }
         }
      }
   dataregs[r1].t.b1 = tempvar;
   linetrue = dataregs[r1].t.b0 & TIMER_DONE_BIT ? TRUE : FALSE;
   program_ip += 4;
   }

void ctf()
   {
   tempvar = linetrue;  /* need to save this before we change it */
   if (!linetrue)
      {
      if (!(dataregs[r1].t.b0 & TIMER_DONE_BIT))  /* if not already done */
         {
         dataregs[r1].t.b0 = TIMER_ENABLED_BIT | TIMER_ACTIVE_BIT;
         if (dataregs[r1].t.b1)  /* TRUE -> FALSE transition */
            {
            if (++dataregs[r1].t.w >= *(WORD *)&program[program_ip+2])
               dataregs[r1].t.b0 = TIMER_ENABLED_BIT | TIMER_DONE_BIT;
            }
         }
      }
   dataregs[r1].t.b1 = tempvar;
   linetrue = dataregs[r1].t.b0 & TIMER_DONE_BIT ? TRUE : FALSE;
   program_ip += 4;
   }

void tmt()
   {
   if (linetrue)
      {
      if (!(dataregs[r1].t.b0 & TIMER_DONE_BIT))  /* if not already done */
         {
         if (!(dataregs[r1].t.b0 & TIMER_ENABLED_BIT))  /* initting timer */
            dataregs[r1+1].v = tick_counter;
         dataregs[r1].t.b0 = TIMER_ENABLED_BIT | TIMER_ACTIVE_BIT;
         if ((tick_counter - dataregs[r1+1].v) >= (DWORD)(*(WORD *)&program[program_ip+2]))
                        /* timer expired */
            {
            dataregs[r1].t.b0 = TIMER_ENABLED_BIT | TIMER_DONE_BIT;
            }
         }
      }
   else
      {
      dataregs[r1].t.b0 = 0;  /* all flags off */
      }
   linetrue = dataregs[r1].t.b0 & TIMER_DONE_BIT ? TRUE : FALSE;
   program_ip += 4;
   }

void tmf()
   {
   if (!linetrue)
      {
      if (!(dataregs[r1].t.b0 & TIMER_DONE_BIT))  /* if not already done */
         {
         if (!(dataregs[r1].t.b0 & TIMER_ENABLED_BIT))  /* initting timer */
            dataregs[r1+1].v = tick_counter;
         dataregs[r1].t.b0 = TIMER_ENABLED_BIT | TIMER_ACTIVE_BIT;
         if ((tick_counter - dataregs[r1+1].v) >= (DWORD)(*(WORD *)&program[program_ip+2]))
                        /* timer expired */
            {
            dataregs[r1].t.b0 = TIMER_ENABLED_BIT | TIMER_DONE_BIT;
            }
         }
      }
   else
      {
      dataregs[r1].t.b0 = 0;  /* all flags off */
      }
   linetrue = dataregs[r1].t.b0 & TIMER_DONE_BIT ? TRUE : FALSE;
   program_ip += 4;
   }

void rtmt()
   {
   if (linetrue)
      {
      if (!(dataregs[r1].t.b0 & TIMER_DONE_BIT))  /* if not already done */
         {
         if (!(dataregs[r1].t.b0 & TIMER_ENABLED_BIT))  /* initting timer */
            dataregs[r1+1].v = tick_counter - dataregs[r1].t.w;
         dataregs[r1].t.b0 = TIMER_ENABLED_BIT | TIMER_ACTIVE_BIT;
         if ((tick_counter - dataregs[r1+1].v) >= (DWORD)(*(WORD *)&program[program_ip+2]))
                        /* timer expired */
            {
            dataregs[r1].t.b0 = TIMER_ENABLED_BIT | TIMER_DONE_BIT;
            }
         }
      }
   else
      {
      if (dataregs[r1].t.b0 & TIMER_ENABLED_BIT)
         {
         dataregs[r1].t.w = tick_counter - dataregs[r1+1].v;
         dataregs[r1].t.b0 &= ~TIMER_ENABLED_BIT;  /* flag off */
         }
      }
   linetrue = dataregs[r1].t.b0 & TIMER_DONE_BIT ? TRUE : FALSE;
   program_ip += 4;
   }

void rtmf()
   {
   if (!linetrue)
      {
      if (!(dataregs[r1].t.b0 & TIMER_DONE_BIT))  /* if not already done */
         {
         if (!(dataregs[r1].t.b0 & TIMER_ENABLED_BIT))  /* initting timer */
            dataregs[r1+1].v = tick_counter - dataregs[r1].t.w;
         dataregs[r1].t.b0 = TIMER_ENABLED_BIT | TIMER_ACTIVE_BIT;
         if ((tick_counter - dataregs[r1+1].v) >= (DWORD)(*(WORD *)&program[program_ip+2]))
                        /* timer expired */
            {
            dataregs[r1].t.b0 = TIMER_ENABLED_BIT | TIMER_DONE_BIT;
            }
         }
      }
   else
      {
      if (dataregs[r1].t.b0 & TIMER_ENABLED_BIT)
         {
         dataregs[r1].t.w = tick_counter - dataregs[r1+1].v;
         dataregs[r1].t.b0 &= ~TIMER_ENABLED_BIT;  /* flag off */
         }
      }
   linetrue = dataregs[r1].t.b0 & TIMER_DONE_BIT ? TRUE : FALSE;
   program_ip += 4;
   }

void reset()
   {
   if (linetrue)
      dataregs[r1].v = 0;  /* reset flags & accum */
   program_ip += 2;
   }

void invert()
   {
   linetrue = !linetrue;
   program_ip++;
   }

void (*optable[LAST_OPCODE+1])() =
   {
   nop,           /* 0x00: NOP */
   nop,           /* 0x01: BREAK, implement later */
   setlinetrue,   /* 0x02: set linetrue to TRUE */
   pushlinetrue,  /* 0x03: push linetrue to init stack */
   orbranch,      /* 0x04: push linetrue to result stack, copy (not pop) linetrue from init stack */
   orpoplinetrue, /* 0x05: or linetrue with top of result stack, pop initstack and result stack */
   inp,           /* 0x06: linetrue = input */
   ninp,          /* 0x07: linetrue = !input */
   out,           /* 0x08: output = linetrue */
   nout,          /* 0x09: output = !linetrue */
   lout,          /* 0x0a: if linetrue set output */
   uout,          /* 0x0b: if linetrue clear output */
   add,           /* 0x0c: if linetrue dest = R1 + R2 */
   addb,          /* 0x0d: if linetrue dest = R1 + BVAL */
   sub,           /* 0x0e: if linetrue dest = R1 - R2 */
   subb,          /* 0x0f: if linetrue dest = R1 - BVAL */
   mul,           /* 0x10: if linetrue dest = R1 * R2 */
   mulb,          /* 0x11: if linetrue dest = R1 * BVAL */
   div,           /* 0x12: if linetrue dest = R1 / R2 */
   divb,          /* 0x13: if linetrue dest = R1 / BVAL */
   mod,           /* 0x14: if linetrue dest = R1 % R2 */
   modb,          /* 0x15: if linetrue dest = R1 % BVAL */
   and,           /* 0x16: if linetrue dest = R1 & R2 */
   andb,          /* 0x17: if linetrue dest = R1 & BVAL */
   or,            /* 0x18: if linetrue dest = R1 | R2 */
   orb,           /* 0x19: if linetrue dest = R1 | BVAL */
   xor,           /* 0x1a: if linetrue dest = R1 ^ R2 */
   xorb,          /* 0x1b: if linetrue dest = R1 ^ BVAL */
   not,           /* 0x1c: if linetrue R1 = ~R1 */
   neg,           /* 0x1d: if linetrue R1 = -R1 */
   equ,           /* 0x1e: linetrue = R1 == R2 */
   equb,          /* 0x1f: linetrue = R1 == BVAL */
   neq,           /* 0x20: linetrue = R1 != R2 */
   neqb,          /* 0x21: linetrue = R1 != BVAL */
   grt,           /* 0x22: linetrue = R1 > R2 */
   grtb,          /* 0x23: linetrue = R1 > BVAL */
   geq,           /* 0x24: linetrue = R1 >= R2 */
   geqb,          /* 0x25: linetrue = R1 >= BVAL */
   les,           /* 0x26: linetrue = R1 < R2 */
   lesb,          /* 0x27: linetrue = R1 < BVAL */
   leq,           /* 0x28: linetrue = R1 <= R2 */
   leqb,          /* 0x29: linetrue = R1 <= BVAL */
   mov,           /* 0x2a: if linetrue dest = R1 */
   movd,          /* 0x2b: if linetrue dest = DVAL */
   jmp,           /* 0x2c: if linetrue goto label */
   osr,           /* 0x2d: if previous scan was FALSE and linetrue, set linetrue for one scan */
   ctt,           /* 0x2e: counter of FALSE -> TRUE transitions */
   nop,           /* 0x2f: not implmenented, reserved */
   ctf,           /* 0x30: counter of TRUE -> FALSE transitions */
   nop,           /* 0x31: not implemented, reserved */
   tmt,           /* 0x32: timer of TRUE duration */
   nop,           /* 0x33: not implemented, reserved */
   tmf,           /* 0x34: timer of FALSE duration */
   nop,           /* 0x35: not implemented, reserved */
   rtmt,          /* 0x36: retentive timer of TRUE duration */
   nop,           /* 0x37: not implemented, reserved */
   rtmf,          /* 0x38: retentive timer of FALSE duration */
   nop,           /* 0x39: not implemented, reserved */
   reset,         /* 0x3a: if linetrue reset timer/counter */
   invert         /* 0x3b: linetrue = !linetrue */
   };

void process_ladder()
   {
   while (runflag)  /* or until BREAK */
      {
      PET_WATCHDOG();
      r1 = program[program_ip+1];
      r2 = program[program_ip+2];
      (*optable[program[program_ip]])();
      if (debugflag)
         {
         if (run_nodecount)
            {
            if (--run_nodecount == 0)
               {
               runflag = FALSE;
               }
            }
         }
      if (program_ip == program_len)  /* end of ladder */
         {
         program_ip = 0;
         break;
         }
      }
   }

update_runstop_status()
   {
   lasthardrun = hardrun;
   if ((hardrun == RUNSTOP_RUN) && (softrun))
      {
      runflag = TRUE;
      }
   else
      {
      runflag = FALSE;
      }
   }

read_eeprom_program()
   {
   WORD p_l, i, chksum;

   p_l = eeprom_program_len;
   if ((p_l) && (p_l <= MAX_PROGRAM))  /* could be valid... */
      {
      for (chksum = 0, i = 0; i < p_l; i++)
         {
         program[i] = eeprom_program[i];
         chksum += program[i];
         }
      if (chksum == eeprom_program_checksum)
         {
         program_len = p_l;
         program_valid = TRUE;
         softrun = TRUE;  /* at powerup, assume software run permission */
         }
      }
   }

void save_program()
   {
   WORD i, chksum;

   if (program_valid)
      {
      eeprom_program_len = program_len;
      for (chksum = 0, i = 0; i < program_len; i++)
         {
         eeprom_program[i] = program[i];
         chksum += program[i];
         }
      eeprom_program_checksum = chksum;
      }
   }

void check_message_type()
   {
   incoming_len = (WORD)message[3] + ((WORD)message[4] << 8);
   if (message[2] > LAST_MSG)
      message_status = MSG_STATUS_IGNORE;
   if (message[2] == MSG_PROGRAM)
      {
      p_len = 0;
      if ((incoming_len < 3) || (incoming_len > MAX_PROGRAM + 2))
         message_status = MSG_STATUS_IGNORE;
      else
         {
         runflag = FALSE;
         softrun = FALSE;
         program_valid = FALSE;
         program_ip = 0;
         debugflag = FALSE;
         message_status = MSG_STATUS_PROGRAM;
         update_runstop_status();
         init_logic();
         }
      }
   else
      if (incoming_len > 16)
         message_status = MSG_STATUS_IGNORE;
   incoming_len += message_ptr;
   }

sendack()
   {
   xmit_char(MSG_BYTE0);
   xmit_char(MSG_BYTE1);
   xmit_char(MSG_ACK);
   xmit_char(0);
   xmit_char(0);
   }

sendstatus()
   {
   xmit_char(linetrue);
   xmit_char(runflag);
   xmit_char(debugflag);
   xmit_char(program_ip & 0xff);
   xmit_char(program_ip >> 8);
   xmit_char(program_valid);
   xmit_char(program_len & 0xff);
   xmit_char(program_len >> 8);
   }

sendregs()
   {
   WORD i;

   sendstatus();
   for (i = 0; i < PLC_MEMORY_LOCS; i++)
      {
      xmit_wait(dataregs[i].bytes[0]);
      xmit_wait(dataregs[i].bytes[1]);
      xmit_wait(dataregs[i].bytes[2]);
      xmit_wait(dataregs[i].bytes[3]);
      }
   }

void process_message()
   {
   WORD i;
   WORD chksum;

   switch (message[2])
      {
      case MSG_RUN:
         if (program_valid)
            {
            softrun = TRUE;
            sendack();
            update_runstop_status();
            }
         break;
      case MSG_STOP:
         softrun = FALSE;
         sendack();
         update_runstop_status();
         break;
      case MSG_GETREGS:
         sendregs();
         break;
      case MSG_NUMSTEPS:
         if ((program_valid) && (message_ptr == 7))
            {
            run_nodecount = (WORD)message[5] + ((WORD)message[6] << 8);
            if (run_nodecount)
               {
               debugflag = TRUE;
               sendack();
               }
            }
         break;
      case MSG_PROGRAM:
         p_len -= 2;
         for (chksum = 0, i = 0; i < p_len; i++)
            chksum += program[i];
         if (((chksum & 0xff) == program[i])
               && ((chksum >> 8) == program[i+1]))
            {
            program_len = p_len;
            program_valid = TRUE;
            sendack();
            save_program();
            }
         break;
      case MSG_PROGSTATUS:
         sendstatus();
         break;
      case MSG_ACK:
         break;
      }
   }

void main(void)
   {
   system_init();

   xmit_str("TinyPLC V0.01A 9/13/06\r\n");
   program_valid = FALSE;
   read_eeprom_program();
   init_logic();
   lasthardrun =
   hardrun = RUNSTOP_SWITCH;
   update_runstop_status();

   while (1)
      {
      if (rx_buf_in)
         {
         if (program_timer == 0)  /* start of incoming */
            program_timer = 10;
         if (program_abort)
            {
            program_abort = FALSE;
            program_timer = 0;
            rx_buf_in = 0;
            rx_buf_out = 0;
            message_ptr = 0;
            incoming_len = 0xffff;  /* invalid! */
            message_status = MSG_STATUS_WAITING;
            }
         else
            {
            DI();
            while (rx_buf_out != rx_buf_in)
               {
               if (message_status == MSG_STATUS_PROGRAM)
                  {
                  if (p_len != MAX_PROGRAM + 2)
                     program[p_len++] = rx_buf[rx_buf_out];
                  }
               else
                  {
                  if ((message_status != MSG_STATUS_IGNORE) && (message_ptr < MAX_MSG_SIZE))
                     message[message_ptr] = rx_buf[rx_buf_out];
                  }
               message_ptr++;
               rx_buf_out = (rx_buf_out == RXBUF_SIZE - 1) ? 0 : rx_buf_out + 1;
               if ((message_ptr == 1) && (message[0] != MSG_BYTE0))
                  message_ptr = 0;
               if ((message_ptr == 2) && (message[1] != MSG_BYTE1))
                  message_ptr = 0;
               if (message_ptr == 5)  /* 55 AA <cmd> <len-low> <len-high> */
                  check_message_type();
               }
            rx_buf_in = 0;
            rx_buf_out = 0;
            EI();
            if (message_ptr >= incoming_len)  /* EOM or error */
               {
               if (message_ptr == incoming_len)
                  process_message();
               program_timer = 0;
               program_abort = FALSE;
               message_ptr = 0;
               message_status = MSG_STATUS_WAITING;
               incoming_len = 0xffff;  /* invalid! */
               }
            }
         }
      PET_WATCHDOG();
      if (program_abort)
         {
         program_abort = FALSE;
         message_ptr = 0;
         incoming_len = 0xffff;  /* invalid! */
         }
      if (++status_led_counter > status_led_pwm)
         STATUS_LED = STATUS_LED_ON;
      else
         STATUS_LED = STATUS_LED_OFF;
      if ((hardrun = RUNSTOP_SWITCH) != lasthardrun)
         update_runstop_status();
      if (program_valid)
         {
         if (program_ip == 0)  /* if about to start the ladder */
            gather_inputs();
         process_ladder();
         if (program_ip == 0)  /* if we just finished the ladder */
            service_outputs();
         }
      }  /* while (1) */
   }
